<?php

namespace Drupal\drealty_reconciler;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\State\StateInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides logic to sync RETS inactive data between RETS and Drupal.
 */
class DrealtyReconcilerService {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Constructs a FileUploadHandler object.
   *
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel factory.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack.
   */
  public function __construct(StateInterface $state, EntityTypeManagerInterface $entityTypeManager, MessengerInterface $messenger, LoggerChannelFactoryInterface $logger_factory, RequestStack $requestStack) {
    $this->state = $state;
    $this->entityTypeManager = $entityTypeManager;
    $this->messenger = $messenger;
    $this->logger = $logger_factory->get('drealty_reconcile');
    $this->requestStack = $requestStack;
  }

  /**
   * Clears the queue out.
   */
  public function clearQueue() {
    \Drupal::queue('drealty_reconciler')->deleteQueue();
  }

  /**
   * Retrieves the current state of the queue.
   *
   * @return mixed
   *   The Drupal queue for Reconciler.
   */
  public static function getQueue() {
    return \Drupal::queue('drealty_reconciler');
  }

  /**
   * Processes Reconciler queue item to unpublish items or delete.
   *
   * @param array $item
   *   The Drupal queue item.
   */
  public function processQueueItem(array $item) {
    $fetcher = \Drupal::service('drealty.fetcher');

    if (empty($item['rets_key']) && $item['delete']) {
      // @todo Delete entity right away, it's already unpublished (verified??).
    }
    elseif (!empty($item['rets_key'])) {
      $db_entity = $fetcher->entityExists($item['entity_type'], $item['rets_key'], NULL, $item['class'], $item['bundle']);

      if ($db_entity) {
        $connection = $this->entityTypeManager->getStorage('drealty_connection')
          ->load($item['conid']);
        $class = $this->entityTypeManager->getStorage('drealty_class')
          ->load($item['class']);
        $resource = $class->get('resource');

        $listing = FALSE;
        $key_field = $connection->getResources()[$resource]['keyfield'];
        $q = $fetcher->_getClassQuery($class);
        $q[] = "({$key_field}={$item['rets_key']})";
        $query = implode(',', $q);

        $options = [
          'Format' => 'COMPACT-DECODED',
          'Limit' => 1,
          'Count' => 0,
        ];

        if (!$db_entity->isPublished() && $item['active'] === 0 && $item['delete']) {
          $db_entity->delete();
          $this->logger->debug("Deleted inactive entity @type: (MLS: @mls ID: @code)", [
            '@type' => $item['entity_type'] . '__' . $item['bundle'],
            "@mls" => $item['rets_key'],
            "@code" => $item['id'],
          ]);
        }
        elseif ($fetcher->connectRetsServer($connection)) {
          $search = $fetcher->SearchQuery($resource, $class->get('systemname'), $query, $options, FALSE);
          $error_info = $search->getError();
          $list_item = $search->first();

          if (!$search || !empty($error_info['code']) || empty($list_item)) {
            // Check RETS error - if connection issue, listing might still
            // exist (chap. #7.8):
            // https://www.nar.realtor/retsorg.nsf/retsproto1.7d6.pdf
            if (!empty($error_info['code']) && $error_info['code'] != '20201') {
              $this->logger->error('Search for UniqueID # @mls listing (Drupal ID # @id) was INTERRUPTED. RETS ERROR: @error', [
                '@mls' => $item['rets_key'],
                '@id' => $item['id'],
                '@error' => $error_info['text'],
              ]);
              return FALSE;
            }
            $this->logger->debug('Search for UniqueID # @key listing was unsuccessful.', [
              '@key' => $item['rets_key'],
            ]);
          }
          else {
            $listing = !empty($list_item) ? $list_item->toArray() : FALSE;
          }

          try {
            if (empty($listing)) {
              if ($item['delete']) {
                $db_entity->delete();
                $this->logger->debug("Deleted inactive entity @type: (MLS: @mls ID: @code)", [
                  '@type' => $item['entity_type'],
                  "@mls" => $item['rets_key'],
                  "@code" => $item['id'],
                ]);
              }
              else {
                $db_entity->set('rets_reconciled', \Drupal::time()
                  ->getRequestTime());
                $db_entity->setUnpublished();
                $db_entity->save();
              }
            }
            elseif (!empty($listing) && is_array($listing)) {
              $field_mappings = $fetcher->findDrealtyMappedFields($item['conid'], $item['class'], $item['entity_type'], $item['bundle']);
              if (!empty($field_mappings)) {
                $mapping = $field_mappings['rets_status']['drealty_mapping'];
                if (!empty($mapping['rets_status__value'])) {
                  $f = $mapping['rets_status__value']['field'];
                  if (!empty($f) && !empty($listing[$f])) {
                    $db_entity->set('rets_status', $listing[$f]);
                  }
                }
                // @todo Decide if we need to update "rets_modify__value" field.
              }
              $db_entity->set('rets_reconciled', \Drupal::time()
                ->getRequestTime());
              $db_entity->setPublished();
              $db_entity->save();
            }
          }
          catch (\Exception $e) {
            $this->logger->error($e->getMessage(), []);
          }
        }
      }
    }
  }

}
