<?php

namespace Drupal\drealty\Commands;

use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\State\StateInterface;
use Drush\Commands\DrushCommands;

/**
 * DRealty drush commands.
 */
class DrealtyCommands extends DrushCommands {

  /**
   * Date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatter
   */
  protected $dateFormatter;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * DrealtyCommands constructor.
   *
   * @param \Drupal\Core\Datetime\DateFormatter $dateFormatter
   *   Date formatter service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager service.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state.
   */
  public function __construct(DateFormatter $dateFormatter, EntityTypeManagerInterface $entityTypeManager, StateInterface $state) {
    parent::__construct();
    $this->dateFormatter = $dateFormatter;
    $this->entityTypeManager = $entityTypeManager;
    $this->state = $state;
  }

  /**
   * Import mapped RETS Classes with fields data (exclude Media, etc.).
   *
   * @param array $options
   *   Array of options as described below.
   *
   * @command rets-import
   * @usage drush rets-import
   *   Import ALL data from scratch, regardless Modification timestamp.
   * @usage drush rets-import --increment=TRUE
   *   Import data starting from certain Modification timestamp recorded when
   *   it was last imported.
   * @aliases ri
   *
   * @option increment Get MLS updates incrementally (since last updated
   *   timestamp).
   */
  public function rets_import(array $options = ['increment' => self::OPT]) {
    // Set the time out to infinite.
    set_time_limit(0);
    $increment = FALSE;

    if (!empty($options['increment'])) {
      $increment = TRUE;
    }
    \Drupal::service('drealty.fetcher')->run_import('resources', $increment);
  }

  /**
   * Process any RETS items remaining/stuck in the Drush queue.
   *
   * @command rets-process-queued-properties
   * @aliases rpqp
   */
  public function rets_process_queued_properties(array $options = []) {
    // Set the time out to infinite.
    set_time_limit(0);
    \Drupal::service('drealty.fetcher')->process_results();
  }

  /**
   * Retrieves Media Objects for all the entities from RETS based on specified
   * Search configuration. Generates Drupal queue for processing.
   *
   * @param array $options
   *   Array of options as described below.
   *
   * @command rets-import-media
   * @usage drush rets-import-media
   *   Queue all RETS entities for Media Objects processing.
   * @usage drush rets-import --increment=TRUE
   *   Import Media Objects for only those entities from RETS that have been
   * modified within time period configured in
   *   admin/drealty/drealty-import/config.
   * @aliases rim
   *
   * @option increment Get MLS updates incrementally (since last updated
   *   timestamp).
   */
  public function rets_import_media(array $options = ['increment' => self::OPT]) {
    set_time_limit(0);
    $increment = FALSE;

    if (!empty($options['increment'])) {
      $increment = TRUE;
    }
    \Drupal::service('drealty.fetcher')->run_import('media', $increment);
  }

  /**
   * Process any Media Objects that are queued up in the system.
   *
   * @command rets-process-media
   * @aliases rpm
   */
  public function rets_process_media() {
    set_time_limit(0);
    \Drupal::service('drealty.fetcher')->import_media();
  }

  /**
   * Process any Media Objects that are queued up in the system.
   *
   * @param string $entity_type
   *   Entity type to process.
   *
   * @command rets-purge-inactive
   * @aliases rpi
   *
   * @option bundle Entity bundle.
   * @option class RETS Class (Drupal config ID format
   *   "conn_id.resource_id.class_id").
   */
  public function rets_purge_inactive($entity_type, array $options = [
    'bundle' => self::OPT,
    'class' => self::OPT,
  ]) {
    // @todo finish it out:
    set_time_limit(0);
    if (!empty($options)) {
      $storage = $this->entityTypeManager->getStorage($entity_type);

      $entity_query = \Drupal::entityQuery($entity_type)
        ->accessCheck(FALSE);
      if (!empty($options['class'])) {
        $entity_query->condition('rets_class', [$options['class']], 'IN');
      }
      if (!empty($options['bundle'])) {
        $entity_query->condition($storage->getEntityType()
          ->getKey('bundle'), [$options['bundle']], 'IN');
      }
    }
    /*
     * else {
      $srvc = \Drupal::service('drealty.fetcher');
      $mapped_entities = $srvc->findClassesMappedEntities();
      foreach ($mapped_entities as $entity_type => $bundles) {
        $entity_type_def = \Drupal::entityTypeManager()->getDefinition($entity_type);
        $entity_query = \Drupal::entityQuery($entity_type)
          ->condition($entity_type_def->getKey('status'), 0)
          ->condition('rets_class', $classes_config, 'IN')
          ->accessCheck(FALSE);
        $deactivate_set = $entity_query->execute();

        foreach ($bundles as $bundle) {
          $field = FieldConfig::loadByName($entity_type, $bundle, 'rets_reconciled');
          if (empty($field)) {
            $field = FieldConfig::create([
              'field_name' => 'rets_reconciled',
              'entity_type' => $entity_type,
              'bundle' => $bundle,
              'label' => 'RETS Last Reconciled',
              'required' => TRUE,
              'translatable' => FALSE,
              'description' => 'The last time this entity was checked against to exist there.',
            ]);
            $field->save();
          }
        }
      }
    }
   */
    $purge_set = $entity_query->execute();
  }

}
