<?php

namespace Drupal\drealty;

use Drupal\Component\Utility\Html;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Config\Entity\ConfigEntityListBuilder;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\drealty\Entity\ConnectionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Url;

/**
 * Provides a listing of connections.
 */
class ConnectionListBuilder extends ConfigEntityListBuilder {

  /**
   * The entity storage class for the 'drealty_class' entity type.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $classStorage;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * The route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs an ConnectionListBuilder object.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param \Drupal\Core\Entity\EntityStorageInterface $storage
   *   The entity storage class.
   * @param \Drupal\Core\Entity\EntityStorageInterface $class_storage
   *   The entity storage class for the 'drealty_class' entity type.
   */
  public function __construct(EntityTypeInterface $entity_type, EntityStorageInterface $storage, EntityTypeManagerInterface $entity_type_manager, Request $request, RouteMatchInterface $route_match) {
    parent::__construct($entity_type, $storage);

    $this->entityTypeManager = $entity_type_manager;
    $this->classStorage = $entity_type_manager->getStorage('drealty_class');
    $this->request = $request;
    $this->routeMatch = $route_match;

  }

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity_type.manager')->getStorage($entity_type->id()),
      $container->get('entity_type.manager'),
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('current_route_match')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultOperations(EntityInterface $entity) {
    /** @var \Drupal\Core\Config\Entity\ConfigEntityInterface $entity */
    $operations = parent::getDefaultOperations($entity);
    // $type = $entity->getEntityTypeId();
    if (\Drupal::currentUser()
      ->hasPermission('administer drealty connections') && $entity->getEntityTypeId()) {

      $operations['view_connection'] = [
        'title' => t('Manage'),
        'url' => Url::fromRoute('entity.drealty_connection.view_page', [
          'drealty_connection' => $entity->id(),
        ]),
        'weight' => 0,
      ];

      // If connection is "enabled", it means MLS credentials were correct and
      // resources can be updated.
      if ($entity->status()) {
        $operations['update_resources'] = [
          'title' => t('Update resources'),
          'url' => Url::fromRoute('entity.drealty_connection.update_resources', [
            'drealty_connection' => $entity->id(),
          ]),
          'weight' => 50,
        ];
      }
      return $operations;
    }
    // If ($entity->access('update') && $entity->hasLinkTemplate('edit-form')) {
    //      $operations['edit'] = [
    //        'title' => $this->t('Edit'),
    //        'weight' => 10,
    //        'url' => $this->ensureDestination($entity->toUrl('edit-form')),
    //      ];
    //    }
    //    if ($entity->access('delete') && $entity->hasLinkTemplate('delete-form')) {
    //      $operations['delete'] = [
    //        'title' => $this->t('Delete'),
    //        'weight' => 100,
    //        'url' => $this->ensureDestination($entity->toUrl('delete-form')),
    //      ];
    //    }.
    return $operations;
  }

  /**
   * {@inheritdoc}
   */
  public function buildHeader() {
    return [
      'type' => $this->t('Type'),
      'title' => $this->t('Name'),
      'status' => $this->t('Status'),
    ] + parent::buildHeader();
  }

  /**
   * {@inheritdoc}
   */
  public function buildRow(EntityInterface $entity) {
    /** @var \Drupal\Core\Config\Entity\ConfigEntityInterface $entity */
    $row = parent::buildRow($entity);
    $status = $entity->status();
    $row = [
      'data' => [
        'type' => [
          'data' => $entity instanceof ConnectionInterface ? $this->t('Connection') : $this->t('Class'),
          'class' => ['drealty-resource-type'],
        ],
        'title' => [
          'data' => [
            '#markup' => $entity->label(),
          ],
          // 'data' => [
          //              '#type' => 'link',
          //              '#title' => $entity->label(),
          //            ] + $entity->toUrl('canonical')->toRenderArray(),
          'class' => ['drealty-resource-title'],
        ],
        'status' => [
          'data' => $status ? $this->t('Enabled') : $this->t('Disabled'),
          'class' => ['drealty-resource-status'],
        ],
        'operations' => $row['operations'],
      ],
      'title' => $this->t('ID: @name', ['@name' => $entity->id()]),
      'class' => [
        Html::cleanCssIdentifier($entity->getEntityTypeId() . '-' . $entity->id()),
        $status ? 'drealty-resource-list-enabled' : 'drealty-resource-list-disabled',
        $entity instanceof ConnectionInterface ? 'drealty-resource-list-connection' : 'drealty-resource-list-class',
      ],
    ];

    $description = $entity->get('description');
    if ($description) {
      $row['data']['title']['data']['#suffix'] = '<div class="description">' . $description . '</div>';
    }

    // If ($status
    //      && $entity instanceof ConnectionInterface) {
    //      $row['data']['status']['data'] = $this->t('Active');
    //      $row['class'][] = 'color-error';
    //    }.
    return $row;
  }

  /**
   * {@inheritdoc}
   */
  public function render() {
    $entity_groups = $this->loadGroups();
    $list['#type'] = 'container';
    $list['#attached']['library'][] = 'drealty/drealty.admin_css';

    if (!empty($entity_groups['connections'])) {
      $list['connections'] = [
        '#type' => 'table',
        '#header' => $this->buildHeader(),
        '#rows' => [],
        '#empty' => '',
        '#attributes' => [
          'id' => 'drealty-entity-list',
          'class' => [
            'drealty-entity-list',
            'drealty-entity-list--connections-with-classes',
          ],
        ],
      ];
      foreach ($entity_groups['connections'] as $connection_groups) {
        /** @var \Drupal\Core\Config\Entity\ConfigEntityInterface $entity */
        foreach ($connection_groups as $entity) {
          $list_builder = $this->entityTypeManager->getListBuilder($entity->getEntityTypeId());
          $list['connections']['#rows'][$entity->getEntityTypeId() . '.' . $entity->id()] = $this->buildRow($entity);
        }
      }
    }

    // Output the list of classes without a connection separately.
    if (!empty($entity_groups['lone_classes'])) {
      $list_builder = $this->entityTypeManager->getListBuilder('drealty_class');
      $list['lone_classes']['heading']['#markup'] = '<h3>' . $this->t('DRealty Classes mapped to this bundle') . '</h3>';
      $list['lone_classes']['table'] = [
        '#type' => 'table',
        '#header' => $list_builder->buildHeader(),
        '#rows' => [],
        '#attributes' => [
          'id' => 'drealty-entity-list',
          'class' => [
            'drealty-entity-list',
            'drealty-entity-list--unattached-classes',
          ],
        ],
      ];

      foreach ($entity_groups['lone_classes'] as $entity) {
        $list['lone_classes']['table']['#rows'][$entity->id()] = $list_builder->buildRow($entity);
      }
    }
    // Elseif (!$list['connections']['#rows']) {
    //      $list['connections']['#empty'] = $this->t('There are no connections or classes defined.');
    //    }.
    return $list;
  }

  /**
   * Loads connections and classes, grouped by connections.
   *
   * @return \Drupal\Core\Config\Entity\ConfigEntityInterface[][]
   *   An associative array with two keys:
   *   - connections: All available connections, each followed by all classes
   *     attached to it.
   *   - lone_classes: All classes that aren't attached to any connection.
   */
  public function loadGroups() {
    $entity_type_id = $this->routeMatch->getParameter('entity_type_id');
    $bundle_entity_type_id = $this->routeMatch->getParameter('bundle_entity_type_id');
    $bundle = $this->routeMatch->getParameter($bundle_entity_type_id);
    $classes = [];
    if ($bundle) {
      $settings = $bundle->getThirdPartySetting('drealty', 'mapping');
      $connections = [];
      foreach ($settings as $conn_id => $setting) {
        $connections[] = $conn_id;
        // $t = array_values($setting);
        //        $t1 = array_column($t, 'class');
        //        $t3 = array_column(array_values($setting), 'class');
        $classes = array_merge($classes, array_column(array_values($setting), 'class'));
      }
      $values = (!empty($connections)) ? ['id' => $connections] : NULL;
      if (!empty($values)) {
        $connections = $this->storage->loadByProperties($values);
        $classes = $this->classStorage->loadByProperties(['id' => $classes]);
        // $values = (!empty($classes)) ? ['id' => $classes] : NULL;
        //        $classes = $this->classStorage->loadByProperties($values);
        $bundle_classes = $classes;
      }
    }
    else {
      $connections = $this->storage->loadMultiple();
      // /** @var \Drupal\drealty\Entity\ConnectionInterface[] $classes */
      //      $classes = $this->classStorage->loadMultiple();
      $bundle_classes = [];
    }

    $this->sortByStatusThenAlphabetically($connections);
    if (!empty($classes)) {
      $this->sortByStatusThenAlphabetically($bundle_classes);
    }

    $connection_groups = [];
    foreach ($connections as $connection) {
      $connection_group = [];
      // $connection_group = [
      //        'connection.' . $connection->id() => $connection,
      //      ];
      if (empty($bundle)) {
        $connection_group = [
          'connection.' . $connection->id() => $connection,
        ];
      }
      elseif ($bundle && !empty($classes)) {
        // Foreach ($classes as $class) {
        //          $connection_group['class.' . $class->id()] = $class;
        //        }.
        // /** @var \Drupal\drealty\Entity\ResourceInterface[] $resources */
        //        $resources = $connection->getResources();
        //        if (!empty($resources)) {
        //          $resource_storage = $this->entityTypeManager->getStorage('drealty_resource');
        //          foreach ($resources as $resource) {
        //            if (!($resource instanceof ResourceInterface)) {
        //              $resource = $resource_storage->load($resource['id']);
        //            }
        //
        //            /** @var \Drupal\drealty\Entity\DrealtyClassInterface[] $resource_classes */
        //            if (!empty($resource) && !empty($resource_classes = $resource->getClasses())) {
        //              if (!empty($bundle_classes)) {
        //                $bundle_classes_ids = array_keys($bundle_classes);
        //                foreach ($resource_classes as $key => $resource_class) {
        //                  if (!in_array($resource_class->id(), $bundle_classes_ids)) {
        //                    unset($resource_classes[$key]);
        //                  }
        //                }
        //              }
        //              foreach ($resource_classes as $class) {
        //                $connection_group['class.' . $class->id()] = $class;
        //                // Remove this class from $classes so it will finally only contain those
        //                // classes not belonging to any connection.
        //                unset($classes[$class->id()]);
        //              }
        //            }
        //          }
        //        }
      }

      $connection_groups['connection.' . $connection->id()] = $connection_group;
    }

    // If (empty($bundle)) {
    //      return [
    //        'connections' => $connection_groups,
    //        //          'lone_classes' => $classes,
    //      ];
    //    }
    //    else {
    //      return [
    //        'lone_classes' => $classes,
    //      ];
    //    }.
    return [
      'connections' => $connection_groups,
      'lone_classes' => $classes,
    ];
  }

  /**
   * Sorts an array of entities by status and then alphabetically.
   *
   * Will preserve the key/value association of the array.
   *
   * @param \Drupal\Core\Config\Entity\ConfigEntityInterface[] $entities
   *   An array of config entities.
   */
  protected function sortByStatusThenAlphabetically(array &$entities) {
    uasort($entities, function (ConfigEntityInterface $a, ConfigEntityInterface $b) {
      if ($a->status() == $b->status()) {
        return strnatcasecmp($a->label(), $b->label());
      }
      else {
        return $a->status() ? -1 : 1;
      }
    });
  }

}
