<?php

namespace Drupal\drealty\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;
use PHRETS\Versions\RETSVersion;

/**
 * Connection form.
 *
 * @property \Drupal\drealty\Entity\ConnectionInterface $entity
 */
class ConnectionForm extends EntityForm {

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#maxlength' => 255,
      '#default_value' => $this->entity->label(),
      '#description' => $this->t('Label for the connection.'),
      '#required' => TRUE,
      '#size' => 60,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#machine_name' => [
        'exists' => ['\Drupal\drealty\Entity\Connection', 'load'],
        'source' => ['label'],
      ],
      '#disabled' => !$this->entity->isNew(),
      '#description' => $this->t('A unique machine-readable name for this DRealty Connection. It must only contain lowercase letters, numbers, and underscores.'),
    ];

    $form['status'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enabled'),
      '#default_value' => $this->entity->status(),
    ];

    $form['description'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description'),
      '#default_value' => $this->entity->get('description'),
      '#description' => $this->t('Description of the connection.'),
    ];

    $form['login_url'] = [
      '#type' => 'textfield',
      '#title' => t('Login URL'),
      '#description' => $this->t('Login URL given to you by your RETS provider. <em>i.e. (http://demo.crt.realtors.org:6103/rets/login)</em>'),
      '#required' => TRUE,
      '#size' => 80,
    ];
    if (!empty($this->entity->getLoginUrl())) {
      $form['login_url']['#default_value'] = $this->entity->getLoginUrl();
    }
    $form['username'] = [
      '#type' => 'textfield',
      '#title' => t('Username'),
      '#size' => 40,
      '#required' => TRUE,
    ];
    if (!empty($this->entity->getUsername())) {
      $form['username']['#default_value'] = $this->entity->getUsername();
    }
    $form['password'] = [
      '#type' => 'textfield',
      '#title' => t('Password'),
      '#size' => 40,
      '#required' => TRUE,
    ];
    if (!empty($this->entity->getPassword())) {
      $form['password']['#default_value'] = $this->entity->getPassword();
    }

    $form['ua_string'] = [
      '#type' => 'textfield',
      '#title' => t('User Agent String'),
      '#description' => t('A User Agent String.'),
      '#required' => TRUE,
      '#size' => 80,
    ];
    if (!empty($this->entity->getUaString())) {
      $form['ua_string']['#default_value'] = $this->entity->getUaString();
    }
    $form['ua_password'] = [
      '#type' => 'textfield',
      '#title' => t('User Agent Password'),
      '#description' => t('(Leave blank if you do not have one).'),
      '#required' => FALSE,
      '#size' => 40,
      '#default_value' => $this->entity->getUaPassword(),
    ];
    if (!empty($this->entity->getUaPassword())) {
      $form['ua_password']['#default_value'] = $this->entity->getUaPassword();
    }

    $rets_versions = (new RETSVersion())->getValidVersions();
    $form['version'] = [
      '#type' => 'select',
      '#title' => t('RETS Version'),
      '#options' => array_combine($rets_versions, $rets_versions),
      '#required' => TRUE,
    ];
    if (!empty($this->entity->getVersion())) {
      $form['version']['#default_value'] = $this->entity->getVersion();
    }

    $form['use_interealty_auth'] = [
      '#type' => 'checkbox',
      '#title' => t('Use Interealty Authentication'),
      '#default_value' => $this->entity->isUseInterealtyauth(),
    ];
    $form['force_basic_auth'] = [
      '#type' => 'checkbox',
      '#title' => t('Force Basic Authentication'),
      '#default_value' => $this->entity->isForceBasicauth(),
    ];
    $form['use_compression'] = [
      '#type' => 'checkbox',
      '#title' => t('Use Compression'),
      '#default_value' => $this->entity->isUseCompression(),
    ];

    $form['disable_encoding_fix'] = [
      '#type' => 'checkbox',
      '#title' => t('Disable Encoding Fix'),
      '#default_value' => $this->entity->isDisableEncodingfix(),
    ];

    $form['debug_mode'] = [
      '#type' => 'checkbox',
      '#title' => t('Enable Debug Mode'),
      '#description' => t('Writes phRets debug messages to drealty_debug_log.txt in the private files folder.'),
      '#default_value' => $this->entity->isDebugMode(),
    ];

    $form['import_modified_span'] = [
      '#type' => 'select',
      '#title' => 'Modified Timestamp span',
      '#description' => 'When querying RETS for incremental updates, what should the date threshold be to query for? Smaller values will yield fewer results. Ideally, this value should be set to 3 days.',
      '#options' => [
        '4' => '4 hours',
        '6' => '6 hours',
        '12' => '12 hours',
        '1' => '1 day',
        '3' => '3 days',
        '5' => '5 days',
        '7' => '1 week',
        '14' => '2 weeks',
        '30' => '1 month',
        '90' => '3 months',
        '180' => '6 months',
        '365' => '1 year',
        '730' => '2 years',
        '0' => 'Any time',
      ],
      '#default_value' => $this->entity->get('import_modified_span') ? $this->entity->get('import_modified_span') : 3,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(&$form, FormStateInterface $form_state) {
    if ($form_state->getValue('status')) {
      $valid = $this->entity->connect();
      if (!$valid) {
        $this->messenger()
          ->addError('Verify ' . strtoupper($this->entity->label()) . ' configuration as Connection with MLS failed to be established.');
        $form_state->setValue('status', FALSE);
      }
    }

    parent::submitForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $result = parent::save($form, $form_state);
    $message_args = ['%label' => $this->entity->label()];
    $message = $result == SAVED_NEW
      ? $this->t('Created new connection %label.', $message_args)
      : $this->t('Updated connection %label.', $message_args);
    $this->messenger()->addStatus($message);

    $connect_config = \Drupal::configFactory()
      ->getEditable('drealty.drealty_connection.' . $this->entity->id());
    if ($this->entity->status() && $connect_config && empty($connect_config->get('resources'))) {

      $batch = [
        'title' => t('Fetching Metadata from MLS'),
        'init_message' => t('Preparing for download...'),
        'progress_message' => t('Fetching metadata'),
        'error_message' => t('The sync process has encountered an error.'),
        'finished' => 'Drupal\drealty\Form\ConnectionUpdateResourcesForm::batchFinished',
      ];

      if (!empty($resources = $this->entity->getRemoteResources())) {
        $this->entity->setResources([]);
        $batch['operations'][] = [
          'Drupal\drealty\Form\ConnectionUpdateResourcesForm::batchProcess',
          [$this->entity, $resources, count($resources)],
        ];
      }
      batch_set($batch);

    }
    else {
      $form_state->setRedirectUrl($this->entity->toUrl('collection'));
    }

    return $result;
  }

  /**
   * Batch callback; initialize the number of updated resources.
   */
  public static function batchStart(&$context) {
    $context['results']['updates'] = 0;
  }

  /**
   * Common batch processing callback for all operations.
   *
   * Required to load our include the proper batch file.
   */
  public static function batchProcess($connection, $resources, $ttl, &$context) {
    if (!isset($context['sandbox']['current'])) {
      $context['sandbox']['count'] = 0;
      $context['sandbox']['current'] = 0;
      $context['sandbox']['total'] = $ttl;
    }
    // Get the total amount of items to process.
    if (!isset($context['sandbox']['total'])) {
      $context['sandbox']['total'] = $ttl;
      // If there are no entities to update, then stop immediately.
      if (!$context['sandbox']['total']) {
        $context['finished'] = 1;
        return;
      }
    }

    if ($context['sandbox']['total']) {
      $conn_upd = (empty($context['sandbox']['connection']) && $connection) ? $connection : $context['sandbox']['connection'];
      if (!empty($resources[$context['sandbox']['current']])) {
        $context['sandbox']['connection'] = $connection->getFetcher()
          ->setConnectionMetadataConfig($conn_upd, [$resources[$context['sandbox']['current']]]);
        $context['sandbox']['count'] += 1;
        $context['message'] = 'Processed ' . $resources[$context['sandbox']['current']]->systemname . ' MLS resource. Updated ' . $context['sandbox']['count'] . ' out of ' .
          $context['sandbox']['total'] . ' total.';
        $context['sandbox']['current'] += 1;
        $context['results']['updates'] += 1;
      }
    }

    if (!empty($context['sandbox']['total'])) {
      $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
      if ($context['finished'] == 1) {
        $context['message'] = 'Processed ' . $context['sandbox']['total'] . ' Resource from MLS.';
      }
    }
  }

  /**
   * Batch finished callback.
   */
  public static function batchFinished($success, $results, $operations) {
    if ($success) {
      if ($results['updates']) {
        \Drupal::messenger()->addStatus(\Drupal::translation()
          ->formatPlural($results['updates'], 'Refreshed metadata for 1 Resource.', 'Refreshed metadata for @count Resources.'));
      }
      else {
        \Drupal::messenger()->addStatus(t('No Resource records to process.'));
      }
    }
    else {
      $error_operation = reset($operations);
      \Drupal::messenger()
        ->addStatus(t('An error occurred while processing @operation with arguments : @args', [
          '@operation' => $error_operation[0],
          '@args' => print_r($error_operation[0], TRUE),
        ]));
    }
  }

}
