<?php

namespace Drupal\drealty\Form;

use Drupal\Core\Entity\EntityConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\drealty\RetsFetcherInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a confirmation form for resending order receipts.
 */
class ConnectionUpdateResourcesForm extends EntityConfirmFormBase {

  /**
   * RETS Fetcher.
   *
   * @var \Drupal\drealty\RetsFetcherInterface
   */
  protected $fetcher;

  /**
   * Constructs a new OrderReceiptResendForm object.
   *
   * @param \Drupal\drealty\RetsFetcherInterface $fetcher
   *   RETS fetcher.
   */
  public function __construct(RetsFetcherInterface $fetcher) {
    $this->fetcher = $fetcher;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('drealty.fetcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('Are you sure you want to update resources from remote for connection %label?', [
      '%label' => $this->entity->label(),
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function getConfirmText() {
    return $this->t('Update resources');
  }

  /**
   * {@inheritdoc}
   */
  public function getCancelUrl() {
    return $this->entity->toUrl('collection');
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $batch = [
      'title' => t('Updating Metadata from MLS'),
      'init_message' => t('Preparing for update...'),
      'progress_message' => t('Updating metadata'),
      'error_message' => t('The sync process has encountered an error.'),
      'finished' => 'Drupal\drealty\Form\ConnectionUpdateResourcesForm::batchFinished',
    ];

    if (!empty($resources = $this->entity->getRemoteResources())) {
      $this->entity->setResources([]);
      $batch['operations'][] = [
        'Drupal\drealty\Form\ConnectionUpdateResourcesForm::batchProcess',
        [$this->entity, $resources, count($resources)],
      ];
    }
    batch_set($batch);
    $this->messenger()
      ->addMessage($this->t('Metadata information was updated from MLS.'));
  }

  /**
   * Batch callback; initialize the number of updated resources.
   */
  public static function batchStart(&$context) {
    $context['results']['updates'] = 0;
  }

  /**
   * Common batch processing callback for all operations.
   *
   * Required to load our include the proper batch file.
   */
  public static function batchProcess($connection, $resources, $ttl, &$context) {
    if (!isset($context['sandbox']['current'])) {
      $context['sandbox']['count'] = 0;
      $context['sandbox']['current'] = 0;
      $context['sandbox']['total'] = $ttl;
    }
    // Get the total amount of items to process.
    if (!isset($context['sandbox']['total'])) {
      $context['sandbox']['total'] = $ttl;
      // If there are no entities to update, then stop immediately.
      if (!$context['sandbox']['total']) {
        $context['finished'] = 1;
        return;
      }
    }

    if ($context['sandbox']['total']) {
      $conn_upd = (empty($context['sandbox']['connection']) && $connection) ? $connection : $context['sandbox']['connection'];
      if (!empty($resources[$context['sandbox']['current']])) {
        $context['sandbox']['connection'] = $connection->getFetcher()
          ->setConnectionMetadataConfig($conn_upd, [$resources[$context['sandbox']['current']]]);
        $context['sandbox']['count'] += 1;
        $context['message'] = 'Processed ' . $resources[$context['sandbox']['current']]->systemname . ' MLS resource. Updated ' . $context['sandbox']['count'] . ' out of ' .
          $context['sandbox']['total'] . ' total.';
        $context['sandbox']['current'] += 1;
        $context['results']['updates'] += 1;
      }
    }

    if (!empty($context['sandbox']['total'])) {
      $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
      if ($context['finished'] == 1) {
        $context['message'] = 'Processed ' . $context['sandbox']['total'] . ' Resource from MLS.';
      }
    }
  }

  /**
   * Batch finished callback.
   */
  public static function batchFinished($success, $results, $operations) {
    if ($success) {
      if ($results['updates']) {
        \Drupal::messenger()->addStatus(\Drupal::translation()
          ->formatPlural($results['updates'], 'Refreshed metadata for 1 Resource.', 'Refreshed metadata for @count Resources.'));
      }
      else {
        \Drupal::messenger()->addStatus(t('No Resource records to process.'));
      }
    }
    else {
      $error_operation = reset($operations);
      \Drupal::messenger()
        ->addStatus(t('An error occurred while processing @operation with arguments : @args', [
          '@operation' => $error_operation[0],
          '@args' => print_r($error_operation[0], TRUE),
        ]));
    }
  }

}
