<?php

namespace Drupal\drealty\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\drealty\QueryTypeInterface;
use Drupal\drealty\RetsFetcherInterface;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use PHRETS\Exceptions\CapabilityUnavailable;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * DRealty class form.
 *
 * @property \Drupal\drealty\Entity\DrealtyClassInterface $entity
 */
class DrealtyClassForm extends EntityForm {

  use MessengerTrait;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * The route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * RETS Fetcher service.
   *
   * @var \Drupal\drealty\RetsFetcherInterface
   */
  protected $fetcher;

  /**
   * Constructs a new WorkspaceActivateForm.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Request.
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   Route matcher.
   * @param \Drupal\drealty\RetsFetcherInterface $fetcher
   *   RETS Fetcher service.
   */
  public function __construct(Request $request, RouteMatchInterface $route_match, RetsFetcherInterface $fetcher) {
    $this->request = $request;
    $this->routeMatch = $route_match;
    $this->fetcher = $fetcher;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('current_route_match'),
      $container->get('drealty.fetcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {

    $form = parent::form($form, $form_state);

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#maxlength' => 255,
      '#default_value' => $this->entity->label(),
      '#description' => $this->t('Label for DRealty class.'),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#machine_name' => [
        'exists' => ['\Drupal\class\Entity\DrealtyClass', 'load'],
        'replace_pattern' => '[^a-zA-Z0-9_.]+',
      ],
      '#disabled' => !$this->entity->isNew(),
    ];

    $form['status'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enabled'),
      '#default_value' => $this->entity->status(),
    ];

    $form['description'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description'),
      '#default_value' => $this->entity->get('description'),
      '#description' => $this->t('Description of the drealty class.'),
    ];

    $connection = $this->entityTypeManager->getStorage('drealty_connection')
      ->load($this->entity->getConnectionId());
    $entity_type = $this->routeMatch->getParameter('entity_type_id');

    $form_state->set('connection', $connection);
    $form_state->set('entity_type', $entity_type);
    $form_state->set('class', $this->entity);

    $form['required'] = [
      '#type' => 'details',
      '#title' => $this->t('Required Fields'),
      '#open' => TRUE,
    ];

    $form['required']['rets_id'] = [
      '#autocomplete_route_name' => 'drealty.autocomplete.class_fields',
      '#autocomplete_route_parameters' => [
        'drealty_connection' => $this->entity->getConnectionId(),
        'drealty_class' => $this->entity->id(),
      ],
      '#type' => 'textfield',
      '#title' => $this->t('Id Field'),
      '#description' => $this->t('A unique ID field for the RETS resource, for Properties this is typically the MLS ID.'),
      '#required' => TRUE,
      '#default_value' => $this->getIdField('id'),
    ];

    $form['required']['rets_key'] = [
      '#autocomplete_route_name' => 'drealty.autocomplete.class_fields',
      '#autocomplete_route_parameters' => [
        'drealty_connection' => $this->entity->getConnectionId(),
        'drealty_class' => $this->entity->id(),
      ],
      '#type' => 'textfield',
      '#title' => $this->t('Key Field'),
      '#description' => $this->t('The unique key field for the RETS resource as defined by the RETS provider. Can be the same as the ID field, however some RETS providers have seperate fields for the Key and the ID. Images are also typically based off of the Key field rather than the ID field.'),
      '#required' => TRUE,
      '#default_value' => $this->getIdField('key'),
    ];
    $form['required']['rets_status'] = [
      '#autocomplete_route_name' => 'drealty.autocomplete.class_fields',
      '#autocomplete_route_parameters' => [
        'drealty_connection' => $this->entity->getConnectionId(),
        'drealty_class' => $this->entity->id(),
      ],
      '#type' => 'textfield',
      '#title' => $this->t('Status Field'),
      '#description' => $this->t('The RETS Listing status field. If your server supports offset, then this field must be mapped in order to build the default RETS query.'),
      '#required' => FALSE,
      '#default_value' => $this->getIdField('status'),
    ];

    $form['required']['rets_modified_field'] = [
      '#autocomplete_route_name' => 'drealty.autocomplete.class_fields',
      '#autocomplete_route_parameters' => [
        'drealty_connection' => $this->entity->getConnectionId(),
        'drealty_class' => $this->entity->id(),
      ],
      '#type' => 'textfield',
      '#title' => $this->t('Modified Field'),
      '#description' => $this->t('The RETS DateTime field which holds the date when the resource was last modified in the RETS system, e.g. LastChangeTimestamp or MatrixModifiedDT. Will be utilized during import requests to eliminate processing of properties which have not changed.'),
      '#required' => TRUE,
      '#default_value' => $this->getIdField('modified'),
    ];

    $form['feed'] = [
      '#type' => 'details',
      '#title' => $this->t('RETS Import Options'),
      '#collapsible' => FALSE,
      '#open' => TRUE,
      '#description' => $this->t('Use the following settings to configure the import options for the RETS feed.'),
    ];

    $skip_status = FALSE;
    $query_types = [
//      QueryTypeInterface::DEFAULT => 'Default (Server must support offset. Relies on RETS Status Field.)',
//      QueryTypeInterface::PRICE => 'Offset <strong>Not</strong> Supported (Query based on price field)',
      QueryTypeInterface::RETS_KEY => 'Offset <strong>Not</strong> Supported (Query based on RETS Key)',
      QueryTypeInterface::MANUAL => 'Manually supplied Query (Server must support offset)',
    ];

    $opts = [];

    $form['feed']['query_options'] = [
      '#type' => 'details',
      '#title' => 'Query Options',
    ];

    $form['feed']['query_options']['query_type'] = [
      '#type' => 'radios',
      '#title' => $this->t('Query Type'),
      '#default_value' => $this->entity->get('query_type') !== NULL ? $this->entity->get('query_type') : QueryTypeInterface::DEFAULT,
      '#options' => $query_types,
      '#description' => $this->t('Please select the type of query operation to use.'),
    ];
    if (!$skip_status && !empty($opts)) {
      $form['feed']['query_options']['status'] = [
        '#type' => 'select',
        '#title' => $this->t('Status Field'),
        '#options' => $opts,
        '#default_value' => explode(',', $this->entity->get('status_values')),
        '#description' => $this->t('Select the appropriate values to limit the search query.'),
        '#multiple' => TRUE,
        '#states' => [
          'visible' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::DEFAULT]],
          'required' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::DEFAULT]],
        ],
      ];
    }
    else {
      $form_state->set('status_null', TRUE);
    }
    $form['feed']['query_options']['status_override_query'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Query to use'),
      '#rows' => 4,
      '#description' => $this->t('Must be a valid DMQL query. (specs found <a href="http://retsdoc.onconfluence.com/display/rets172/7.7+Query+language" target="_blank">here.</a> A good tutorial can be found <a href="http://www.flexmls.com/support/rets/tutorials/dmql/" target="_blank">here.</a> You may also find that <a href="http://www.retsmd.com" target="_blank">www.retsmd.com</a> is a valuable resource in helping write the query.'),
      '#default_value' => $this->entity->get('override_status_query_text'),
      '#states' => [
        'visible' => [
          ':input[name="query_type"]' => [
            ['value' => QueryTypeInterface::MANUAL],
            ['value' => QueryTypeInterface::PRICE],
          ],
        ],
        'required' => [
          ':input[name="query_type"]' => [
            ['value' => QueryTypeInterface::MANUAL],
            ['value' => QueryTypeInterface::PRICE],
          ],
        ],
      ],
    ];

    $form['feed']['query_options']['check_query_1'] = [
      '#type' => 'submit',
      '#value' => $this->t('Test DMQL Query'),
      '#description' => $this->t('This process may take several minutes.'),
      '#submit' => ['::testManualQuery'],
      '#states' => [
        'visible' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
      ],
    ];
    $form['feed']['query_options']['check_query'] = [
      '#type' => 'submit',
      '#value' => $this->t('Test DMQL Query'),
      '#description' => $this->t('This process my take several minutes.'),
      '#submit' => ['::testManualQuery'],
      '#states' => [
        'visible' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::MANUAL]],
      ],
    ];
    $form['feed']['query_options']['offset_field'] = [
      '#type' => 'entity_autocomplete',
      '#title' => $this->t('Price field to base the offset on'),
      '#description' => $this->t('This field must be a numeric price field. If you are unsure, visit <a href="http://www.retsmd.com" target="_blank">www.retsmd.com</a> to ensure that the field is numeric.'),
      '#default_value' => $this->entity->get('offset_field'),
      '#target_type' => 'drealty_field',
      '#selection_handler' => 'default',
      '#selection_settings' => [
        'connection' => $connection,
        'class' => $this->entity->id(),
      ],
      '#states' => [
        'visible' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
        'required' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
      ],
    ];

    $form['feed']['query_options']['offset_amount'] = [
      '#type' => 'select',
      '#title' => $this->t('The amount to offset by.'),
      '#default_value' => $this->entity->get('offset_amount'),
      '#options' => [
        10000 => '$10,000',
        25000 => '$25,000',
        50000 => '$50,000',
        75000 => '$75,000',
        100000 => '$100,000',
        200000 => '$200,000',
      ],
      '#states' => [
        'visible' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
        'required' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
      ],
    ];

    $form['feed']['query_options']['offset_max'] = [
      '#type' => 'select',
      '#title' => $this->t('The maximum offset amount.'),
      '#default_value' => $this->entity->get('offset_max'),
      '#options' => [
        1000000 => '$1m',
        1500000 => '$1.5m',
        2000000 => '$2m',
        5000000 => '$5m',
        7500000 => '$7.5m',
        10000000 => '$10m',
        999000000 => '$999m',
      ],
      '#states' => [
        'visible' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
        'required' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
      ],
    ];

    $form['feed']['query_options']['chunk'] = [
      '#title' => $this->t('Record chunk size'),
      '#type' => 'select',
      '#options' => array_combine([250, 500, 1000, 1500, 2000, 3000], [
        250,
        500,
        1000,
        1500,
        2000,
        3000,
      ]) + ['0' => $this->t('Unlimited')],
      '#default_value' => $this->entity->get('chunk_size'),
      '#description' => $this->t('When processing incoming records, the system will work with data structures containing this many records simultaneously,
      and cache the rest for later. If updating listings is often failing due to out of memory errors and increasing PHP&rsquo;s memory allocation is not an option,
      try decreasing this value. However, decreasing this value increases the amount of time it will take to update listings. The RETS server may override this value (MAXROWS).'),
      '#states' => [
        'invisible' => [':input[name="query_type"]' => ['value' => QueryTypeInterface::PRICE]],
      ],
    ];

    $obj = $connection->findClassObjects($this->entity);
    if (!empty($obj)) {
      $form['feed']['images']['process_images'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Process media?'),
        '#description' => $this->t('Media objects attached are processed after the main data. It can include photos, attachments, videos.'),
        '#default_value' => $this->entity->get('process_images'),
      ];

      $chunk_opts = [
        '5' => '5',
        '10' => '10',
        '25' => '25',
        '40' => '40',
        '50' => '50',
      ];
      $form['feed']['images']['image_chunk_size'] = [
        '#type' => 'select',
        '#multiple' => FALSE,
        '#title' => $this->t('Media chunk size'),
        '#description' => $this->t('Media objects are downloaded in batches. Resource can have 0 or more media objects associated with it. The chunk size determines how many parent resources are included in a batch. If each resource has 10 images associated with it and the chunk size
            is set to 10 a total of 100 images would potentially be downloaded at one time. This setting is meant to <em>adjust</em> the number of images that are downloaded in a single round trip to the RETS server. The more images downloaded the more
            memory that is used by the system to process them. <strong>If you are experiencing out of memory errors while the system is processing images try decreasing the size.</strong> Default is 25'),
        '#options' => $chunk_opts,
        '#default_value' => $this->entity->get('image_chunk_size') ? $this->entity->get('image_chunk_size') : '25',
        '#states' => [
          'visible' => [':input[name="process_images"]' => ['checked' => TRUE]],
        ],
      ];

      $form['feed']['images']['force_update_media'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Mark all media objects for update?'),
        '#description' => $this->t('Set all media hashes to 0 to force updates on the next run.'),
        '#default_value' => FALSE,
        '#states' => [
          'visible' => [':input[name="process_images"]' => ['checked' => TRUE]],
        ],
      ];
    }

    $form['feed']['force_update_data'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Mark all resources for update?'),
      '#description' => $this->t('Set all entities hashes to 0 to force updates on the next run.'),
      '#default_value' => FALSE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $entity_type_id = $this->routeMatch->getParameter('entity_type_id');
    $bundle_entity_type_id = $this->routeMatch->getParameter('bundle_entity_type_id');
    $bundle_entity = $this->routeMatch->getParameter($bundle_entity_type_id);

    $this->createField($form_state->getValue('rets_id'), 'ID');
    $this->createField($form_state->getValue('rets_key'), 'Key');
    $this->createField($form_state->getValue('rets_status'), 'Status');
    $this->createField($form_state->getValue('rets_modified_field'), 'Modified');
    $this->createField(NULL, 'Class');
    $this->createField(NULL, 'Hash');
    $this->createField(NULL, 'MediaHash');

    $this->entity->set('chunk_size', $form_state->getValue('chunk'));
    $this->entity->set('query_type', $form_state->getValue('query_type'));
    $this->entity->set('override_status_query_text', $form_state->getValue('status_override_query'));

    $exists = (!empty($this->entity->drealty_mapping) && in_array($bundle_entity->id(), $this->entity->drealty_mapping[$entity_type_id]));
    if ($this->entity->status()) {
      if (!$exists) {
        $this->entity->drealty_mapping[$entity_type_id][] = $bundle_entity->id();
      }
    }
    elseif (!$this->entity->status() && $exists) {
      $k = array_search($bundle_entity->id(), $this->entity->drealty_mapping[$entity_type_id]);
      unset($this->entity->drealty_mapping[$entity_type_id][$k]);
      $this->entity->drealty_mapping[$entity_type_id] = array_values($this->entity->drealty_mapping[$entity_type_id]);
    }
    if (isset($this->entity->drealty_mapping[$entity_type_id]) && empty($this->entity->drealty_mapping[$entity_type_id])) {
      unset($this->entity->drealty_mapping[$entity_type_id]);
    }

    if ($form_state->getValue('force_update_data')) {
      $this->updateRetsHashes($entity_type_id, $bundle_entity, $this->entity->id());
    }
    if ($form_state->getValue('force_update_media')) {
      $this->updateRetsHashes($entity_type_id, $bundle_entity, $this->entity->id(), 'media');
    }

    $result = $this->entity->save();

    $message_args = ['%label' => $this->entity->label()];
    $message = ($result == SAVED_NEW)
      ? $this->t('Created new drealty class %label.', $message_args)
      : $this->t('Updated drealty class %label.', $message_args);
    $this->messenger()->addStatus($message);

    $form_state->setRedirect("entity.{$entity_type_id}.drealty_class_list", [$bundle_entity_type_id => $bundle_entity->id()]);

    return $result;
  }

  /**
   * Submit handler for testing manual queries.
   *
   * @see drealty_resource_configure_form()
   */
  public function testManualQuery($form, FormStateInterface $form_state) {
    $this->messenger()->addStatus($this->t("Testing the DMQL Query."));

    /** @var \Drupal\drealty\Entity\Connection $connection */
    $connection = $form_state->get('connection');
    $class = $form_state->get('class');
    $resource = $class->get('resource');
    $query = $form_state->getValue('status_override_query');

    if ($this->fetcher->connectRetsServer($connection)) {
      $options = [
        'Count' => 1,
        'Limit' => 1,
        'Format' => 'COMPACT-DECODED',
      ];

      try {
        $search = $this->fetcher->SearchQuery($resource, $class->get('systemname'), $query, $options, FALSE);

        $this->messenger()
          ->addStatus($this->t("Success. RETS server returned @returned of @total results.", [
            '@returned' => $search->getReturnedResultsCount(),
            '@total' => $search->getTotalResultsCount(),
          ]));
      }
      catch (CapabilityUnavailable $e) {
        watchdog_exception('drealty', $e);
        $this->messenger()
          ->addError($this->t("drealty encountered an error: @message", [
            "@message" => $e->getMessage(),
          ]));
      }
    }

    // Rebuild the form to preserve any changes to fields.
    $form_state->setRebuild(TRUE);
  }

  /**
   * Create specific RETS related fields for storage.
   */
  public function createField($fieldName, $field_type) {
    $names = [
      'Modified' => [
        'field' => 'rets_modify',
        'type' => 'string',
        'locked' => TRUE,
        'translatable' => FALSE,
      ],
      'Class' => [
        'field' => 'rets_class',
        'type' => 'string',
        'locked' => TRUE,
        'translatable' => FALSE,
      ],
      'ID' => [
        'field' => 'rets_id',
        'type' => 'string',
        'locked' => TRUE,
        'translatable' => FALSE,
      ],
      'Key' => [
        'field' => 'rets_key',
        'type' => 'string',
        'locked' => TRUE,
        'translatable' => FALSE,
      ],
      'Status' => [
        'field' => 'rets_status',
        'type' => 'list_string',
        'locked' => FALSE,
        'translatable' => TRUE,
      ],
      'Hash' => [
        'field' => 'rets_hash',
        'type' => 'string',
        'locked' => TRUE,
        'translatable' => FALSE,
      ],
      'MediaHash' => [
        'field' => 'rets_media_hash',
        'type' => 'string',
        'locked' => TRUE,
        'translatable' => FALSE,
      ],
    ];
    $cid = str_replace('.', '__', $this->entity->id());
    $field_name = $names[$field_type]['field'];
    // $field_name = mb_strtolower($this->entity->id() . '_' . $field_type);
    $entity_type_id = $this->routeMatch->getParameter('entity_type_id');
    $bundle_entity_type_id = $this->routeMatch->getParameter('bundle_entity_type_id');
    $bundle_entity = $this->routeMatch->getParameter($bundle_entity_type_id);
    if (empty($fieldStorage = FieldStorageConfig::loadByName($entity_type_id, $field_name))) {
      $new_field = [
        'field_name' => $names[$field_type]['field'],
        'langcode' => \Drupal::languageManager()->getDefaultLanguage()->getId(),
        'entity_type' => $entity_type_id,
        'type' => $names[$field_type]['type'],
        'module' => 'drealty',
        'locked' => $names[$field_type]['locked'],
        'cardinality' => 1,
        'translatable' => $names[$field_type]['translatable'],
        'persist_with_no_fields' => FALSE,
        'custom_storage' => FALSE,
      ];
      $fieldStorage = FieldStorageConfig::create($new_field);
      $fieldStorage->save();
    }
    $field = FieldConfig::loadByName($entity_type_id, $bundle_entity->id(), $field_name);
    if (empty($field)) {
      $field = FieldConfig::create([
        'field_name' => $field_name,
        'entity_type' => $entity_type_id,
        'bundle' => $bundle_entity->id(),
        'label' => 'MLS ' . $field_type,
        'required' => FALSE,
        'translatable' => TRUE,
        'description' => 'MLS ' . $field_type,
      ]);
      $field->save();
    }
    else {
      $mapping = $field->getThirdPartySetting('drealty', 'mapping');
    }
    // Some fields don't need to have mapping and are generated to store Rets
    // info mainly:
    if (!empty($fieldName)) {
      $item = [
        'property' => 'value',
        'class' => $this->entity->id(),
        'className' => $this->entity->label() . " (" . $this->entity->getConnectionId() . ")",
        'field' => $fieldName,
      ];
      // Since these'll be protected text type fields, "$field_name . '__value'"
      // is the logical default mapping for field property:
      $mapping[$this->entity->getConnectionId()][$cid][$field_name . '__value'] = $item;
      $field->setThirdPartySetting('drealty', 'mapping', $mapping);
      $field->save();
    }
  }

  /**
   * Find a field which maps to RETS ID.
   */
  public function getIdField(string $type) {
    $names = [
      'modified' => 'rets_modify',
      'class' => 'rets_class',
      'id' => 'rets_id',
      'key' => 'rets_key',
      'status' => 'rets_status',
      'mediahash' => 'rets_media_hash',
    ];
    $field_name = $names[$type];

    $entity_type_id = $this->routeMatch->getParameter('entity_type_id');
    $bundle_entity_type_id = $this->routeMatch->getParameter('bundle_entity_type_id');
    $bundle_entity = $this->routeMatch->getParameter($bundle_entity_type_id);
    $field = FieldConfig::loadByName($entity_type_id, $bundle_entity->id(), $field_name);

    if ($field) {
      $cid = str_replace('.', '__', $this->entity->id());
      $mapping = $field->getThirdPartySetting('drealty', 'mapping');
      if (!empty($mapping[$this->entity->getConnectionId()][$cid])) {
        $f = reset($mapping[$this->entity->getConnectionId()][$cid]);
        return $f['field'];
      }
    }
    return NULL;
  }

  /**
   * Reset RETS Hash values in the database.
   */
  public function updateRetsHashes($entity_type_id, $bundle_obj, $cid, $type = 'data') {
    $entity_type = $bundle_obj->getEntityType();
    $bundle_obj_id = $bundle_obj->id();
    $field = 'rets_hash';
    if ($type == 'media') {
      $field = 'rets_media_hash';
    }
    \Drupal::database()
      ->query("UPDATE {" . $entity_type_id . "__" . $field . "} as h
        JOIN {" . $entity_type_id . '__rets_class' . "} as c on h.entity_id = c.entity_id
        SET " . $field . "_value = 'update' WHERE c.rets_class_value like '$cid' and h.bundle like '$bundle_obj_id'");

    if (isset($entity_type->getKeys()['revision'])) {
      \Drupal::database()
        ->query("UPDATE {" . $entity_type_id . "_revision__" . $field . "} as h
        JOIN {" . $entity_type_id . '__rets_class' . "} as c on h.entity_id = c.entity_id
        SET " . $field . "_value = 'update' WHERE c.rets_class_value like '$cid' and h.bundle like '$bundle_obj_id'");
    }
    $this->messenger()
      ->addMessage($this->t('Hash values have been reset for @cid class @type.', [
        '@cid' => $cid,
        '@type' => $type,
      ]));
  }

}
