<?php

namespace Drupal\drealty\Form;

use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

/**
 * Form to enable mappings from RETS Classes to Drupal content entities.
 */
class EntitiesForm extends EntityForm {

  /**
   * Select available content entities to map them to DRealty Classes.
   *
   * @inheritDoc
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $form['intro'] = [
      '#type' => 'markup',
      '#markup' => '<p>Here you can select available content entities to map 
     them to DRealty Classes.<br>First, check the entity to enable, and then you\'ll
     be able to configure details for it.</p>',
    ];
    $default_value = [];
    $options = [];

    // Get all available CONTENT type entities from DB (e.g. nodes, block).
    $entity_types = $this->entityTypeManager->getDefinitions();
    $content_entity_types = array_filter($entity_types, function (EntityTypeInterface $entity_type) {
      return $entity_type instanceof ContentEntityTypeInterface;
    });

    /** @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface $bundle_info_service */
    $bundle_info_service = \Drupal::service('entity_type.bundle.info');
    foreach ($content_entity_types as $definition) {
      // Get all available bundles for found CONTENT type entities.
      $bundles = $bundle_info_service->getBundleInfo($definition->id());

      foreach ($bundles as $bundle_id => $bundle_info) {
        $bundle_entity_type_id = $definition->getBundleEntityType();
        if (!$bundle_entity_type_id) {
          continue;
        }
        // Get bundle storage config to find ThirdPartySettings:
        $bundle = $this->entityTypeManager->getStorage($bundle_entity_type_id)
          ->load($bundle_id);
        $id = $definition->id() . '__' . $bundle_id;

        $third_settings = $bundle->getThirdPartySetting('drealty', 'mapping', []);
        $config_id = str_replace('.', '__', $this->entity->id());

        $operations = NULL;

        // If found existing mapping, add configuration links:
        if (!empty($third_settings[$this->entity->getConnectionId()][$config_id])) {
          $default_value[$id] = $id;
          $links['configure'] = [
            'title' => $this->t('Configure'),
            'url' => Url::fromRoute(
              "entity.{$definition->id()}.drealty_class_edit_form",
              [
                $bundle_entity_type_id => $bundle_id,
                'drealty_class' => $this->entity->id(),
              ]
            ),
          ];
          $operations['data'] = [
            '#type' => 'operations',
            '#links' => $links,
          ];
        }
        $item = [
          'entity_type' => $definition->getLabel(),
          'bundle' => $bundle_info['label'],
          'operations' => $operations,
        ];
        $options[$id] = $item;
      }
    }

    if (!empty($options)) {
      $header = [
        'entity_type' => t('Entity Type'),
        'bundle' => t('Bundle'),
        'operations' => t('Operations'),
      ];

      $form['entities'] = [
        '#type' => 'tableselect',
        '#header' => $header,
        '#options' => $options,
        '#js_select' => FALSE,
        '#sticky' => TRUE,
        '#responsive' => TRUE,
        '#default_value' => $default_value,
      ];

      $form['actions']['#type'] = 'actions';
      $form['actions']['submit'] = [
        '#type' => 'submit',
        '#value' => t('Save'),
        '#button_type' => 'primary',
      ];
    }

    return $form;
  }

  /**
   * Form submission.
   *
   * @inheritDoc
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Get selected ('enabled') entities and add mapping config:
    $entities = array_filter($form_state->getValue('entities'));
    foreach ($entities as $id) {
      [$entity_type_id, $bundle] = explode('__', $id);
      $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
      $bundle_entity_type_id = $entity_type->getBundleEntityType();
      $bundle = $this->entityTypeManager->getStorage($bundle_entity_type_id)
        ->load($bundle);

      $settings = $bundle->getThirdPartySetting('drealty', 'mapping', []);
      $config_id = str_replace('.', '__', $this->entity->id());
      $settings[$this->entity->getConnectionId()][$config_id] = [
        'connection' => $this->entity->getConnectionId(),
        'class' => $this->entity->id(),
        'resource' => $this->entity->resource,
      ];
      $bundle->setThirdPartySetting('drealty', 'mapping', $settings);
      $bundle->save();
    }
  }

}
