<?php

namespace Drupal\drealty\Form;

use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\drealty\Entity\DrealtyClass;
use Drupal\drealty\RetsFetcher;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 *
 */
class ImportForm extends FormBase {

  /**
   * Batch Builder.
   *
   * @var \Drupal\Core\Batch\BatchBuilder
   */
  protected $batchBuilder;

  /**
   * @var \Drupal\drealty\RetsFetcher
   */
  protected $fetcher;

  /**
   * BatchForm constructor.
   */
  public function __construct(RetsFetcher $fetcher) {
    $this->batchBuilder = new BatchBuilder();
    $this->fetcher = $fetcher;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('drealty.fetcher')
    );
  }

  /**
   * @inheritDoc
   */
  public function getFormId() {
    return 'drealty_import';
  }

  /**
   * @inheritDoc
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $route_match = $this->getRouteMatch();
    $route_params = $route_match->getParameters();
    $conn_arr = [];

    $entity_type_id = $route_match->getParameter('entity_type_id');
    if (!empty($entity_type_id)) {
      // This is the case on routes like this "/admin/structure/types/manage/listing/drealty/import".
      $form['entity_type_id'] = [
        '#type' => 'hidden',
        '#value' => $entity_type_id,
      ];
      $bundle_entity_type_id = $this->getRouteMatch()
        ->getParameter('bundle_entity_type_id');
      $form['bundle_entity_type_id'] = [
        '#type' => 'hidden',
        '#value' => $bundle_entity_type_id,
      ];

      $bundle_entity = $this->getRouteMatch()
        ->getParameter($bundle_entity_type_id);
      $form['bundle_entity_id'] = [
        '#type' => 'hidden',
        '#value' => $bundle_entity->id(),
      ];
      $third_mapping = $bundle_entity->getThirdPartySetting('drealty', 'mapping');
      $connections = array_keys($third_mapping);
      if (empty($connections)) {
        return $form;
      }
      // @todo make multiple.
      $connection_id = reset($connections);
      $form['connection_id'] = [
        '#type' => 'value',
        '#value' => $connection_id,
      ];
      $connection = \Drupal::entityTypeManager()
        ->getStorage('drealty_connection')
        ->load($connection_id);
    }
    else {
      $connection = $route_match->getParameter('drealty_connection');
      if (!empty($connection) && $connection->status()) {
        $conn_arr[$connection->id()] = $connection;
        $form_state->set('connections', $conn_arr);
      }
    }

    $form['help'] = [
      '#markup' => $this->t('Import MLS data using @connection connection.', [
        '@connection' => $connection->label(),
      ]),
    ];

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['run'] = [
      '#type' => 'submit',
      '#value' => $this->t('Run Import'),
      '#button_type' => 'primary',
    ];

    return $form;
  }

  /**
   * @inheritDoc
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $conn_arr = $form_state->get('connections');
    $classes_config = \Drupal::entityQuery('drealty_class')
      ->condition('status', TRUE)
      ->condition('connection', array_keys($conn_arr), 'IN')
      ->execute();
    if (!empty($classes_config)) {
      $classes_config = \Drupal::entityTypeManager()
        ->getStorage('drealty_class')
        ->loadMultiple($classes_config);
    }

    $this->batchBuilder
      ->setTitle($this->t('Processing'))
      ->setInitMessage($this->t('Initializing.'))
      ->setProgressMessage($this->t('Completed @current of @total.'))
      ->setErrorMessage($this->t('An error has occurred.'));

    $this->batchBuilder->setFile(\Drupal::service('extension.list.module')
      ->getPath('drealty') . '/src/Form/ImportForm.php');

    $this->batchBuilder->addOperation([$this, 'processFullImportItems'], [
      $conn_arr,
      $classes_config,
    ]);
    $this->batchBuilder->setFinishCallback([$this, 'finished']);

    \Drupal::queue('drealty_import')->deleteQueue();
    batch_set($this->batchBuilder->toArray());
  }

  /**
   * Processor for batch operations.
   */
  public function processItems($items, DrealtyClass $class, array &$context) {
    // Elements per operation.
    $limit = 50;

    // Set default progress values.
    if (empty($context['sandbox']['progress'])) {
      $context['sandbox']['progress'] = 0;
      $context['sandbox']['max'] = count($items);
    }

    // Save items to array which will be changed during processing.
    if (empty($context['sandbox']['items'])) {
      $context['sandbox']['items'] = $items;
    }

    $counter = 0;
    if (!empty($context['sandbox']['items'])) {
      // Remove already processed items.
      if ($context['sandbox']['progress'] != 0) {
        array_splice($context['sandbox']['items'], 0, $limit);
      }

      foreach ($context['sandbox']['items'] as $item) {
        if ($counter != $limit) {
          $this->processItem($item, $this->t('Populating queue for @class class and entity KEY = @key...', [
            '@class' => $class->id(),
            '@key' => $item['remote_id'],
          ]));

          $counter++;
          $context['sandbox']['progress']++;

          $context['message'] = $this->t('Now processing node :progress of :count', [
            ':progress' => $context['sandbox']['progress'],
            ':count' => $context['sandbox']['max'],
          ]);

          // Increment total processed item values. Will be used in finished
          // callback.
          $context['results']['processed'] = $context['sandbox']['progress'];
        }
      }
    }

    // If not finished all tasks, we count percentage of process. 1 = 100%.
    if ($context['sandbox']['progress'] != $context['sandbox']['max']) {
      $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
    }
  }

  /**
   * Processor for batch operations.
   */
  public function processFullImportItems($conn_arr, $classes_config, array &$context) {
    $to_process = array_values($classes_config);
    if (!isset($context['sandbox']['current'])) {
      $context['sandbox']['count'] = 0;
      $context['sandbox']['current'] = 0;
      $context['sandbox']['total'] = count($classes_config);
    }
    // Get the total amount of items to process.
    if (!isset($context['sandbox']['total'])) {
      $context['sandbox']['total'] = count($classes_config);
      // If there are no entities to update, then stop immediately.
      if (!$context['sandbox']['total']) {
        $context['finished'] = 1;
        return;
      }
    }

    if ($context['sandbox']['total']) {
      $curr_class = $to_process[$context['sandbox']['current']];
      $curr_connection = $conn_arr[$curr_class->getConnectionId()];
      $class_mappings = $curr_class->drealty_mapping;
      $class_mappings['rets_key'] = $curr_connection->getResources()[$curr_class->resource]['keyfield'];
      $mls_items = $this->fetcher->ProcessRetsClass($curr_connection, $curr_class->resource, $curr_class, $class_mappings, 'resources', TRUE);
      if (!empty($mls_items)) {
        $queue = \Drupal::queue('drealty_import');
        foreach ($mls_items as $item) {
          $queue->createItem($item);
        }
      }
      // \Drupal::state()->set('drealty.' . $curr_class->id() . '.lastupdate', time());
      $context['sandbox']['count'] += 1;
      $context['sandbox']['current'] += 1;
      $context['results']['updates'] += 1;
      $context['message'] = 'Populated queue for ' . $curr_class->label() . ' MLS Class (' . $context['sandbox']['count'] . ' out of ' . $context['sandbox']['total'] . ').';
    }

    if (!empty($context['sandbox']['total'])) {
      $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
      if ($context['finished'] == 1) {
        // $context['message'] = 'Processed ' . $context['sandbox']['total'] . ' Classes from MLS.';
      }
    }
  }

  /**
   * Process single item.
   *
   * @param int|string $nid
   *   An id of Node.
   * @param \Drupal\Core\Datetime\DrupalDateTime $date
   *   An object with new published date.
   */
  public function processItem($item, $message) {
    $queue = \Drupal::queue('drealty_import');
    $queue->createItem($item);
    $context['message'] = $message;
    $context['results'][] = $item['remote_id'];
  }

  /**
   * Finished callback for batch.
   */
  public function finished($success, $results, $operations) {
    $message = $this->t('Number of items queued: @count', [
      '@count' => $results['processed'],
    ]);

    $this->messenger()
      ->addStatus($message);
  }

}
