<?php

namespace Drupal\drealty;

use Drupal\Component\Render\PlainTextOutput;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\File\FileSystemInterface;
use Drupal\drealty\Entity\Connection;
use Drupal\drealty\Entity\ConnectionInterface;
use Drupal\drealty\Entity\DrealtyClass;
use Drupal\drealty\Metadata\Field;
use Drupal\drealty\Metadata\Resource;
use Drupal\drealty\Metadata\RetsClass;
use Drupal\drealty\Metadata\RetsObject;
use Drupal\field\Entity\FieldConfig;
use Drupal\file\FileInterface;
use Drupal\media\Entity\Media;
use Drupal\media\Entity\MediaType;
use PHRETS\Configuration as phRetsConfiguration;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Component\Utility\NestedArray;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\State\StateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Fetches a data from RETS server.
 */
class RetsFetcher implements RetsFetcherInterface {

  use DependencySerializationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The logger channel for Drush.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger_drush;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * @var bool
   *   MLS connection established (active session). Usually, session is valid
   * for 1800 seconds.
   */
  protected $isConnected = FALSE;

  /**
   * @var mixed
   *   Timestamp for the current session to timeout.
   */
  protected $sessionTimeout = NULL;

  /**
   * @var \Drupal\drealty\RetsSession
   *   Rets Session.
   */
  protected $rets;

  /**
   * Constructs a FileUploadHandler object.
   *
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel factory.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack.
   */
  public function __construct(StateInterface $state, EntityTypeManagerInterface $entityTypeManager, MessengerInterface $messenger, LoggerChannelFactoryInterface $logger_factory, RequestStack $requestStack) {
    $this->state = $state;
    $this->entityTypeManager = $entityTypeManager;
    $this->messenger = $messenger;
    $this->logger = $logger_factory->get('drealty_fetch');
    $this->logger_drush = $logger_factory->get('drush');
    $this->requestStack = $requestStack;
  }

  /**
   * Destructor for connection.
   */
  public function __destruct() {
    // If we're still connected, someone forgot to disconnect.
    if ($this->isRetsSessionAlive()) {
      // $backtrace = debug_backtrace();
      // $this->logger->warning('Connection to server still active, it was left open. Attempting to disconnect.');
      $this->rets->Disconnect();
    }
  }

  /**
   * Disconnect from the RETS server.
   *
   * @return bool
   *   Whether the session is still alive.
   */
  public function isRetsSessionAlive() {
    return ($this->rets && $this->isConnected && $this->sessionTimeout &&
      $this->sessionTimeout >= time());
  }

  /**
   * {@inheritdoc}
   */
  public function connectRetsServer(ConnectionInterface $connection) {
    try {
      if (!$this->isRetsSessionAlive()) {
        $config = new phRetsConfiguration();
        $config->setLoginUrl($connection->getLoginUrl());
        $config->setUsername($connection->getUsername());
        $config->setPassword($connection->getPassword());

        $config->setRetsVersion($connection->getVersion());
        $config->setUserAgent($connection->getUaString());
        $config->setUserAgentPassword($connection->getUaPassword());
        $auth_method = $connection->isForceBasicauth() ? 'basic' : 'digest';
        $config->setHttpAuthenticationMethod($auth_method);
        // $config->setOption('use_post_method', $query['use_post_method']);
        // $config->setOption('disable_follow_location', $query['disable_follow_location']);
        // Get a session ready using the configuration. Used \PHRETS\Session.
        $this->rets = new RetsSession($config);

        // Make the first request.
        $connect_bulletin = $this->rets->Login();
        if ($connect_bulletin && empty($connect_bulletin->getBody())) {
          $this->isConnected = FALSE;
          $this->sessionTimeout = NULL;
        }
        elseif ($connect_bulletin && !empty($connect_bulletin->getBody())) {
          $this->isConnected = TRUE;
          // Just set it for 25 mins to give some room:
          $this->sessionTimeout = time() + 1500;
        }
      }
      return $this->isRetsSessionAlive();
    }
    catch (\Exception $exception) {
      watchdog_exception('drealty', $exception);
      return FALSE;
    }
  }

  /**
   * Disconnect from the RETS server.
   *
   * @return bool
   */
  public function disconnectRetsServer() {
    if ($this->isRetsSessionAlive()) {
      if ($this->rets->Disconnect()) {
        $this->isConnected = FALSE;
        $this->sessionTimeout = NULL;
        if (function_exists('drush_log')) {
          drush_log("Disconnected from RETS server.");
        }
        return TRUE;
      }
      else {
        // @todo log a message?
      }
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function getRETSResourcesStructured(ConnectionInterface $connection) {
    if ($result = $this->getRemoteResources($connection)) {
      $metadata = [];
      foreach ($result as $resource) {
        $metadata[] = new Resource($connection, $resource);
      }
    }
    return $metadata;
  }

  /**
   * {@inheritdoc}
   */
  public function getRemoteResources(ConnectionInterface $connection) {
    if (!$this->rets) {
      $this->connectRetsServer($connection);
    }
    return $this->rets->GetSystemMetadata()->getResources();
  }

  /**
   * {@inheritdoc}
   */
  public function getRemoteClasses(ConnectionInterface $connection, Resource $resource = NULL) {
    $classes = [];
    $result = $this->getRemoteResources($connection);
    foreach ($result as $remote_resource) {
      if ($resource && $remote_resource->getResourceID() != $resource->systemname) {
        continue;
      }
      $classes[] = $remote_resource->getClasses();
    }
    return $classes;
  }

  /**
   * {@inheritdoc}
   */
  public function getClassesMetadata(ConnectionInterface $connection, Resource $resource = NULL) {
    if ($result = $this->getRemoteClasses($connection, $resource)) {
      $metadata = [];
      foreach ($result as $drealty_classes) {
        foreach ($drealty_classes as $drealty_class) {
          $metadata[] = new RetsClass($connection, $drealty_class);
        }
      }
    }
    return $metadata;
  }

  /**
   * {@inheritdoc}
   */
  public function getRemoteFields(ConnectionInterface $connection, RetsClass $resource = NULL) {
    $fields = [];
    if (!$this->rets) {
      $this->connectRetsServer($connection);
    }
    try {
      $remote_fields = $this->rets->GetTableMetadata('Property', $resource->systemname);
      foreach ($remote_fields as $remote_field) {
        $fields[$remote_field->getSystemName()] = $remote_field;
      }
    }
    catch (\Exception $e) {
      return $fields;
    }
    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function getFields(ConnectionInterface $connection, RetsClass $resource = NULL, $remote = FALSE) {
    $id = $connection->getEntityTypeId() . '_' . $connection->id();
    if ($resource) {
      $id .= '_' . $resource->systemname;
    }
    $id .= '_fields';
    // @todo replace!!!
    $resources = \Drupal::state()->get($id);
    // $resources = NULL;
    if ($resources && !$remote) {
      return $resources;
    }

    $metadata = [];
    if ($result = $this->getRemoteFields($connection, $resource)) {
      foreach ($result as $table) {
        $metadata[] = new Field($connection, $table);
      }
    }
    return $metadata;
  }

  /**
   * {@inheritdoc}
   */
  public function getRetsObjectTypes(Connection $connection, $resource_id = 'Property') {
    if (!$this->rets) {
      $this->connectRetsServer($connection);
    }
    if ($this->rets) {
      $metadata = $this->rets->GetObjectMetadata($resource_id);
      return $metadata;
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   *
   * @param \Drupal\drealty\Entity\Connection $connection
   *   RETS connection.
   * @param array|\Drupal\drealty\Entity\Resource[] $resources
   *   RETS resources.
   */
  public function setConnectionMetadataConfig($connection, $resources = []) {
    $local_resources = [];
    $connect_config = \Drupal::configFactory()
      ->getEditable('drealty.drealty_connection.' . $connection->id());
    if ($connect_config) {
      if (empty($resources)) {
        $resources = $this->getRETSResourcesStructured($connection);
      }
      $classStorage = $this->entityTypeManager->getStorage('drealty_class');
      foreach ($resources as $drealty_resource) {
        $objects_arr = [];
        $arr[$drealty_resource->systemname] = [
          'id' => strtolower($drealty_resource->systemname),
          'label' => $drealty_resource->systemname,
          'description' => $drealty_resource->description,
          'lastupdate' => $drealty_resource->lastupdate,
          'keyfield' => $drealty_resource->keyfield,
          'classes' => [],
        ];

        // Get Resource objects:
        $resource_objects = $this->getRetsObjectTypes($connection, $drealty_resource->systemname);
        if (!empty($resource_objects)) {
          foreach ($resource_objects as $obj) {
            $o = new RetsObject($connection, $obj);
            $objects_arr[$obj->getObjectType()] = $o->valuesArray();
          }
        }

        /** @var \Drupal\drealty\Metadata\RetsClass[] $drealty_classes */
        $drealty_classes = $this->getClassesMetadata($connection, $drealty_resource);
        $classes = [];
        if (!empty($drealty_classes)) {
          foreach ($drealty_classes as $drealty_class) {
            $class_id_data = [
              $connection->id(),
              $drealty_resource->systemname,
              $drealty_class->systemname,
            ];
            $config_id = strtolower(implode('.', $class_id_data));
            $id_machine = str_replace('.', '__', $config_id);

            $arr_classes = [
              'id' => $config_id,
              'label' => $drealty_class->standardname,
              'description' => $drealty_class->description,
              'systemname' => $drealty_class->systemname,
              'standardname' => $drealty_class->standardname,
              'visiblename' => $drealty_class->visiblename,
              'timestampfield' => $drealty_class->timestampfield,
              'connection' => $connection->id(),
              'resource' => $drealty_class->resource,
              'fields' => [],
              'objects' => $objects_arr,
            ];

            if (!empty($drealty_class->systemname)) {
              $arr[$drealty_resource->systemname]['classes'][$id_machine] = $arr_classes;
            }

            $class = $classStorage->load($config_id);
            if (!empty($class)) {
              $class->set('label', $drealty_class->standardname);
              $class->set('description', $drealty_class->description);
              $class->set('systemname', $drealty_class->systemname);
              $class->set('standardname', $drealty_class->standardname);
              $class->set('visiblename', $drealty_class->visiblename);
              $class->set('timestampfield', $drealty_class->timestampfield);
              $class->set('connection', $connection->id());
              $class->set('resource', $drealty_class->resource);
              $class->setFields([]);
              $class->setObjects($objects_arr);
            }
            else {
              $class = DrealtyClass::create($arr_classes);
            }

            $fields = $this->getFields($connection, $drealty_class, TRUE);
            $local_fields = [];
            foreach ($fields as $drealty_field) {
              $field_id_data = [
                $drealty_field->conid,
                $drealty_field->resource,
                $drealty_field->class,
                $drealty_field->systemname,
              ];
              $config_id = strtolower(implode('.', $field_id_data));

              $arr_field = [
                'id' => $config_id,
                'label' => $drealty_field->standardname,
                'systemname' => $drealty_field->systemname,
                'standardname' => $drealty_field->standardname,
                'longname' => $drealty_field->longname,
                'datatype' => $drealty_field->datatype,
                'max_length' => $drealty_field->max_length,
                'field_precision' => $drealty_field->field_precision,
                'interpretation' => $drealty_field->interpretation,
                'lookupname' => $drealty_field->lookupname,
                'class' => $drealty_field->class,
                'lastupdate' => $drealty_field->lastupdate,
              ];
              $field = \Drupal\drealty\Entity\Field::create($arr_field);
              $local_fields[$field->systemname] = $field;
            }
            $class->setFields($local_fields);
            if ($class->id() && !empty($drealty_class->systemname) && !empty($local_fields)) {
              $class->save();
              $classes[] = $class;
            }
          }
        }
        $local_resources += $arr;
      }

      $new_rsc = $connection->getResources() + $local_resources;
      $connection->setResources($new_rsc);
      $connection->save();
    }
    return $connection;
  }

  /**
   * {@inheritdoc}
   */
  public function SearchQuery($resource_id, $class_id, $query = '', $options = [], $recursive = TRUE) {
    $result = $this->rets->Search($resource_id, $class_id, $query, $options, $recursive);
    return $result;
  }

  /**
   * Initiate data import from RETS.
   *
   * @param string $src_type
   *   Processing RETS basic fields (= resources) or Media Objects (= media).
   * @param bool $increment
   *   Whether to implement incremental update.
   *
   * @return bool
   *   Completion status.
   */
  public function run_import($src_type, $increment = FALSE) {
    if ($src_type == 'media' && !$increment) {
      \Drupal::queue('drealty_media')->deleteQueue();
    }
    $connections = $this->entityTypeManager
      ->getStorage('drealty_connection')
      ->loadByProperties(['status' => TRUE]);

    foreach ($connections as $connection) {
      $classes_config = \Drupal::entityQuery('drealty_class')
        ->condition('status', TRUE)
        ->condition('connection', $connection->id(), '=')
        ->execute();
      if (!empty($classes_config)) {
        $classes = \Drupal::entityTypeManager()
          ->getStorage('drealty_class')
          ->loadMultiple($classes_config);
      }

      foreach ($classes as $class) {
        $class_lastupdate = \Drupal::state()
          ->get('drealty.' . $class->id() . '.lastupdate', time());
        $class_mappings = $class->drealty_mapping;
        $class_mappings['rets_key'] = $connection->getResources()[$class->resource]['keyfield'];

        $this->MessageErrorLogger('DRealty Importing class: @class (Last updated: @last, Incremental: @incr, Type: @type)', [
          "@class" => $class->label() . " (" . $class->id() . ")",
          "@last" => $class_lastupdate,
          "@incr" => (string) $increment,
          "@type" => $src_type,
        ], 'both', 'debug');

        if ($src_type == 'resources') {
          \Drupal::state()
            ->set('drealty.' . $class->id() . '.lastupdate_trigger', time());
          $res = $this->ProcessRetsClass($connection, $class->resource, $class, $class_mappings, $src_type, FALSE, $increment);
          \Drupal::state()
            ->set('drealty.' . $class->id() . '.lastupdate', time());
        }
        elseif ($src_type == 'media' && $class->getEnableImageProcessing()) {
          \Drupal::state()
            ->set('drealty.' . $class->id() . '.media_lastupdate_trigger', time() - 60);
          $res = $this->ProcessRetsClass($connection, $class->resource, $class, $class_mappings, $src_type, FALSE, $increment);

          \Drupal::state()
            ->set('drealty.' . $class->id() . '.media_lastupdate', time() - 60);
        }
      }
    }

    unset($connections, $mappings, $classes);
    return TRUE;
  }

  /**
   * Process the current class for import.
   *
   * @param \Drupal\drealty\Entity\Connection $connection
   * @param string $resource_id
   * @param \Drupal\drealty\Entity\DrealtyClass $class
   * @param array $class_mappings
   * @param string $process_type
   *   Processing RETS basic fields (= resources) or Media Objects (= media).
   * @param mixed|bool $return_items
   *   Indicates whether to return array of found items or just queue them up
   *   into $return_items.
   * @param bool $increment
   *   Indicates whether to account for configured ModifiedTimestamp values.
   */
  public function ProcessRetsClass($connection, $resource_id, $class, $class_mappings, $process_type = 'resources', $return_items = FALSE, $increment = FALSE) {
    // $this->queue = DrupalQueue::get('drealty');
    // $this->media_queue = DrupalQueue::get('drealty_media');
    // $this->cron_queue = DrupalQueue::get('drealty_import');
    if (!$increment && $process_type == 'resources') {
      \Drupal::queue('drealty_import')->deleteQueue();
    }

    // Build a list of fields we are going to request from the RETS server.
    $fieldmappings = [];
    $key_field = $class_mappings['rets_key'];
    // Create queue items for processing entity fields and mark the entities which need media updated.
    $query = $this->_getClassQuery($class, $fieldmappings);

    // Create queue items for processing entity fields and mark the entities which need media updated.
    switch ($class->get('query_type')) {
      case QueryTypeInterface::PRICE:
        // @todo check functionality: should it have manual query entry? does it work correctly?
        $this->fetch_listings_offset_not_supported_price($connection, $resource_id, $class, $key_field, $class_mappings, $process_type, $return_items, $increment);
        $process = TRUE;
        break;

      case QueryTypeInterface::MANUAL:
        $process = $this->fetch_listings_offset_supported_default($connection, $resource_id, $class, $query, $key_field, $class_mappings, $process_type, $return_items, $increment);
        break;

      case QueryTypeInterface::RETS_KEY:
        // @todo check functionality: should it have manual query entry? does it work correctly?
        $this->fetch_listings_offset_not_supported_key($connection, $resource_id, $class, $key_field, $class_mappings, $process_type, $return_items, $increment);
        $process = TRUE;
        break;

      case QueryTypeInterface::DEFAULT:
      default:
        // Build the query.
        $statuses = $class->status_values;
        $status_q = "|$statuses";
        $query_field = 'rets_status';
        $query = [];
        $query[] = "{$fieldmappings[$query_field]->systemname}={$status_q}";
        $process = $this->fetch_listings_offset_supported_default($connection, $resource_id, $class, $query, $key_field, $class_mappings, $process_type, $return_items, $increment);
    }

    // At this point we have data waiting to be processed. Need to process the
    // data which will insert/update/delete the listing data as nodes.
    if ($process && $process_type == 'resources') {
      if ($return_items && is_array($process)) {
        return $process;
      }
      else {
        if (\Drupal::queue('drealty_import')->numberOfItems() > 0) {
          $this->MessageErrorLogger("process_results( connection: @connection_name, resource: @resource, class: @class)", [
            "@connection_name" => $connection->id(),
            "@resource" => $resource_id,
            "@class" => $class->label(),
          ], 'drush', 'notice');

          // Process queue items, create entities or update basic fields for existing ones.
          if ($process_type == 'resources') {
            $this->process_results(FALSE);
          }
        }
        else {
          $this->MessageErrorLogger("No items to process. exiting.", [], 'drush', 'notice');
        }
      }
    }
    elseif ($process && $process_type == 'media') {
      $this->MessageErrorLogger("Media queue has been generated.", [], 'drush', 'info');
    }
    else {
      $this->MessageErrorLogger("There was an error returned from RETS of some sort - fetch_listings_offset_supported_default() returned FALSE. No items were queued, none processed.", [], 'drush', 'error');
      return FALSE;
    }

    unset($mappings, $resources, $fieldmappings, $query);
    return TRUE;
  }

  /**
   * Get class query from configuration.
   */
  public function _getClassQuery($class, $fieldmappings = []) {
    $q_array = [];

    switch ($class->get('query_type')) {
      case QueryTypeInterface::PRICE:
      case QueryTypeInterface::RETS_KEY:
        break;

      case QueryTypeInterface::MANUAL:
        $bubbleable_metadata = new BubbleableMetadata();
        $custom_query = \Drupal::token()
          ->replace($class->get('override_status_query_text'), ['drealty-class' => $class], [], $bubbleable_metadata);
        $q_array[] = $custom_query;
        break;

      case QueryTypeInterface::DEFAULT:
      default:
        // Build the query.
        $statuses = $class->status_values;
        $status_q = "|$statuses";
        $query_field = 'rets_status';
        $query = [];
        $query[] = "({$fieldmappings[$query_field]->systemname}={$status_q})";
        $q_array = $query;
    }

    return $q_array;
  }

  /**
   * Process DRealty queue items which contain BASIC mapped fields data.
   *
   * @param bool $process_media
   *   Should we try to pull media for processsing, too?
   */
  public function process_results($process_media = FALSE) {
    try {
      $total = \Drupal::queue('drealty_import')->numberOfItems();
      $i = 1;
      while ($q_item = \Drupal::queue('drealty_import')->claimItem()) {
        $data = $q_item->data;
        $entity_type = $data['entity_type'];
        $bundle = $data['bundle'];
        $storage = $this->entityTypeManager->getStorage($entity_type);

        $db_entity = $this->entityExists($data['entity_type'], $data['remote_id'], NULL, $data['class'], $data['bundle']);
        if (!($db_entity)) {
          $values = [
            $storage->getEntityType()->getKey('bundle') => $bundle,
            $storage->getEntityType()->getKey('status') => 1,
            $storage->getEntityType()->getKey('owner') => 1,
            $storage->getEntityType()
              ->getKey('label') => "MLS Resource data: " . $data['remote_id'],
          ];
          $db_entity = $storage->create($values);
          $db_entity->set('rets_class', $data['class']);
          $db_entity->set('rets_hash', 'update');
          $db_entity->set('rets_media_hash', 'update');
        }

        $up = $this->update_single_listing($db_entity, $data, $process_media);
        if (is_bool($up) && $up) {
          $this->MessageErrorLogger("Saved item @name. [@count of @total]", [
            "@name" => $db_entity->label() . " (Key: " . $db_entity->get('rets_key')->value . ")",
            "@count" => $i,
            "@total" => $total,
          ], 'drush', 'debug');
          \Drupal::queue('drealty_import')->deleteItem($q_item);
        }
        elseif ($up == 'skip') {
          $this->MessageErrorLogger("Skipped item @name. [@count of @total]", [
            "@name" => $db_entity->label() . " (Key: " . $db_entity->get('rets_key')->value . ")",
            "@count" => $i,
            "@total" => $total,
          ], 'drush', 'debug');
          \Drupal::queue('drealty_import')->deleteItem($q_item);
        }
        else {
          $this->MessageErrorLogger("Res returned: @up. Error handling queue item @name. [@count of @total]", [
            "@name" => $db_entity->label() . " (Key: " . $db_entity->get('rets_key')->value . ")",
            "@count" => $i,
            "@total" => $total,
            "@up" => (string) $up,
          ], 'both', 'error');
        }
        $i++;
      }
    }
    catch (\Exception $e) {
      $this->MessageErrorLogger("Error processing queue item. @error. [@count of @total]", [
        "@error" => $e->getMessage(),
        "@count" => $i,
        "@total" => $total,
      ], 'drush', 'error');
    }
  }

  /**
   * Processing Class based on supported Offset while searching.
   *
   * @param \Drupal\drealty\Entity\Connection $connection
   *   RETS connection.
   * @param string $resource_id
   *   RETS resource ID.
   * @param \Drupal\drealty\Metadata\RetsClass $class
   *   RETS class object.
   * @param string $query
   *   Query string to search by.
   * @param string $key_field
   *   Key field for unique mapping.
   * @param array $class_mappings
   *   CLass fields mapping configured to pull the data.
   * @param string $process_type
   *   Processing RETS basic fields (= resources) or Media Objects (= media).
   * @param mixed|bool $return_items
   *   Indicates whether to return array of found items or just queue them up
   *   into $return_items.
   * @param bool $increment
   *   Indicates whether to account for configured ModifiedTimestamp values.
   *
   * @return int
   */
  protected function fetch_listings_offset_supported_default($connection, $resource_id, $class, $query, $key_field, $class_mappings, $process_type, $return_items, $increment) {
    foreach ($class_mappings as $m => $entity) {
      // Cycle through all mapped entities:
      // There's array key to indicate which field is the MLS key field.
      if ($m !== 'rets_key') {
        // Cycle through all of its mapped bundles:
        foreach ($entity as $bundle) {
          $mapping[$m][$bundle] = $this->findDrealtyMappedFields($connection->id(), $class->id(), $m, $bundle);

          // This has to be run for each mapped entity/bundle as config might differ:
          $count = 0;
          $limit = $class->get('chunk_size');
          $queue_items = [];

          if ($limit == 0) {
            $limit = 'NONE';
          }

          if ($this->connectRetsServer($connection)) {
            // Prepare the query.
            $q = implode('),(', $query);
            // $fields = $this->get_fields($connection->conid, $class->cid);.
            $options = [
              'Format' => 'COMPACT-DECODED',
              'Limit' => "$limit",
              'Count' => '1',
              // If 0, there might happen an error: "The offset is bigger than the results.
              // Please limit the offset to less than the last result record. Offset = 669. Results = 668.".
              // PHRETS library needs to receive Count of all records.
              // 'RestrictedIndicator' => 'xxxx',
              // 'Select' => implode(',', $fields),
              // @todo: Search fails when uncommented. Should we just skip it? How resource intensive this is?
            ];

            if ($process_type == 'media') {
              $hash_fields = [];
              if ($class->getEnableImageProcessing()) {
                $obj_mappings = $this->findDrealtyClassMappedObjects($connection->id(), $class->id(), $m, $bundle);
                $time_fields = array_unique(array_column($obj_mappings, 'timestamp'));
                if (!empty($time_fields)) {
                  $hash_fields = array_merge($time_fields, [$key_field]);
                }
              }
              if (!empty($hash_fields)) {
                $options['Select'] = implode(',', $hash_fields);
              }
            }

            // Do the actual search.
            $modified = $connection->get('import_modified_span');
            $mod_field = $mapping[$m][$bundle]['rets_modify']['drealty_mapping']['rets_modify__value'];
            if ($increment && $modified && $modified !== 0 && !empty($mod_field)) {
              // These numbers are hours:
              if (in_array($modified, [2, 4, 6, 12])) {
                $date = substr(\Drupal::service('date.formatter')
                  ->format(strtotime('-' . $modified . ' hours'), 'custom', 'c'), 0, -6) . '+';
              }
              // These numbers are days:
              else {
                $date = \Drupal::service('date.formatter')
                  ->format(strtotime('-' . $modified . ' days'), 'custom', 'Y-m-d') . 'T00:00:00+';
              }
              $search = $this->SearchQuery($resource_id, $class->get('systemname'), $q . ',(' . $mod_field['field'] . '=' . $date . ')', $options);
            }
            else {
              $search = $this->SearchQuery($resource_id, $class->get('systemname'), $q, $options);
            }

            // Loop through the search results.
            if ($search && $search->getReturnedResultsCount() > 0) {
              $rets_items = $search->toArray();
              foreach ($rets_items as $rets_item) {
                if ($process_type == 'resources') {
                  foreach ($class_mappings as $m => $entity) {
                    // Cycle through all mapped entities:
                    if ($m !== 'rets_key') {
                      // Cycle through all of its mapped bundles:
                      foreach ($entity as $bundle) {
                        // Calculate the hash.
                        $rets_item['hash'] = $this->calculate_hash($rets_item, $connection->id(), $class->id(), $mapping[$m][$bundle]);
                        $queue_item = [
                          'resource' => $resource_id,
                          'key_field' => $key_field,
                          'entity_type' => $m,
                          'bundle' => $bundle,
                          'connection' => $connection->id(),
                          'class' => $class->id(),
                          'mapping' => $mapping[$m][$bundle],
                          'result' => $rets_item,
                          'remote_id' => $rets_item[$key_field],
                        ];

                        if ($return_items) {
                          $queue_items[] = $queue_item;
                        }
                        else {
                          \Drupal::queue('drealty_import')
                            ->createItem($queue_item);
                        }
                      }
                    }
                  }
                }
                elseif ($process_type == 'media') {
                  $queue_item = [
                    'conid' => $connection->id(),
                    'rid' => $resource_id,
                    'cid' => $class->id(),
                    'key_field' => $key_field,
                    'key_value' => $rets_item[$key_field],
                    'entity_type' => $m,
                    'bundle' => $bundle,
                    // 'mapping' => $mapping,
                    'rets_media_hash' => $this->calculate_hash($rets_item, $connection->id(), $class->id(), $hash_fields),
                  ];
                  $queue_items[] = $queue_item;
                  \Drupal::queue('drealty_media')->createItem($queue_item);
                }
                $count++;
              }
            }

            if ($error = $search->getError()) {
              $this->MessageErrorLogger("DRealty Search Failed with an error: (Type: @type Code: @code Msg: @text)", [
                "@type" => $error['type'],
                "@code" => $error['code'],
                "@text" => $error['text'],
              ], 'both', 'error');
            }

            // $this->disconnectRetsServer();
            // Do some cleanup.
            // unset($items);
            if ($search->getReturnedResultsCount() > 0) {
              if ($return_items) {
                return $queue_items;
              }
              return TRUE;
            }
            else {
              return FALSE;
            }
          }
          else {
            // $error = $search->getError();
            $this->MessageErrorLogger("Could not connect to RETS server.", [], 'both', 'error');
            return FALSE;
          }
        }
      }
    }

    return FALSE;
  }

  /**
   * Calculate md5 hash on the resulting MLS data.
   *
   * To determine if we need to perform an update on Drupal entity (comparing
   * by stored hash value).
   *
   * @param array $mls_data
   *   Data retrieved from RETS.
   * @param string $conn_id
   *   RETS connection ID.
   * @param string $cid
   *   RETS class identifier.
   * @param array $mapping
   *   RETS to Drupal fields mapping.
   * @param array $forced_items
   *   Additional items to account for hash value generation.
   *
   * @return string
   *   Hash value string.
   */
  protected function calculate_hash($mls_data, $conn_id, $cid, $mapping, $forced_items = []) {

    // $cache = &drupal_static(__FUNCTION__);
    $tmp = '';
    if (stripos($cid, '.') !== FALSE) {
      $cid = str_replace('.', '__', $cid);
    }

    if (!empty($forced_items)) {
      foreach ($forced_items as $i) {
        $tmp .= mb_strtolower(trim($mls_data[$i]));
      }
    }
    else {
      foreach ($mapping as $field) {
        if (!empty($field['drealty_mapping'])) {
          $third = $field['drealty_mapping'];
        }
        elseif (!empty($field['third_party_settings'])) {
          $third = NestedArray::getValue($field['third_party_settings'], [
            'drealty',
            'mapping',
            $conn_id,
            $cid,
          ]);
        }
        elseif (is_string($field) && isset($mls_data[$field])) {
          $tmp .= mb_strtolower(trim($mls_data[$field]));
        }

        if (!empty($third) && is_array($third)) {
          foreach ($third as $item) {
            // "object" is set only on Media field mappings, make sure to skip it:
            if (empty($map_item['object']) && !empty($item['field'])) {
              if (is_string($item['field']) && !empty($mls_data[$item['field']])) {
                $tmp .= mb_strtolower(trim($mls_data[$item['field']]));
              }
              elseif (is_array($item['field'])) {
                $rets_val = [];
                foreach ($item['field'] as $v) {
                  $rets_val[] = $mls_data[$v];
                }
                $rets_val = implode(' ', array_filter($rets_val));
                $tmp .= mb_strtolower(trim($rets_val));
              }
            }
          }
        }
      }
    }

    return md5($tmp);
  }

  /**
   * Update a single entity data with values from RETS.
   */
  public function update_single_listing($db_entity, $rets_item = NULL, $import_media = TRUE) {
    $cid = $db_entity->get('rets_class')->value;
    $entity_type = $db_entity->getEntityTypeId();
    $bundle = $db_entity->bundle();

    // @todo what if rets_item is empty?
    // Skip hash comparison? Will be TRUE when updating single Entity manually.
    $force = FALSE;

    if (empty($cid)) {
      $this->messenger->addMessage(t('Missing info about RETS Class. Sync of RETS data canceled.', []), 'error', FALSE);
      $this->logger->warning('Missing info about RETS Class. Sync of RETS data canceled for @type ID = @id.', [
        '@id' => $db_entity->id(),
        '@type' => $entity_type,
      ]);
      return FALSE;
    }

    try {
      if (!empty($cid)) {
        $class = $this->entityTypeManager->getStorage('drealty_class')
          ->load($cid);
        $conn_id = stristr($cid, '.', TRUE);
        $connection = $this->entityTypeManager->getStorage('drealty_connection')
          ->load($conn_id);
        $field_mappings = $this->findDrealtyMappedFields($conn_id, $cid, $entity_type, $bundle);
      }

      // $db_entity->get('rets_key')->value;
      $key_field = reset($field_mappings['rets_key']['drealty_mapping'])['field'];
      $item_context = [
        'field_mappings' => $field_mappings,
        'class' => $class,
        'connection' => $connection,
        'resource' => $class->get('resource'),
        'key_field' => $key_field,
      ];

      if (!$rets_item) {
        $force = TRUE;
        // If the update was triggered on a Single Listing manually, reset
        // rets_media_hash to ensure fresh data no matter what:
        $db_entity->set('rets_media_hash', 'update');

        $query = "({$key_field}={$db_entity->get($key_field)->value})";
        $params = [
          'Format' => 'COMPACT-DECODED',
          'Limit' => "1",
          'Count' => '0',
        ];

        if ($this->connectRetsServer($connection)) {
          try {
            $search = $this->SearchQuery($class->get('resource'), $class->get('systemname'), $query, $params);
            if (!$search) {
              $this->logger->notice('Search for UniqueID # @key listing (MLS # @mls) was unsuccessful.', [
                '@key' => $db_entity->rets_key->value,
                '@mls' => $db_entity->rets_id->value,
              ]);
              // $this->disconnectRetsServer();
              return FALSE;
            }
            $item = $search->first();
            $item['hash'] = $this->calculate_hash($item, $conn_id, $cid, $field_mappings);
            // $rets->FreeResult($search);
          }
          catch (\Exception $e) {
            $this->logger->error('Error updating entity # @key listing (MLS # @mls). @err.', [
              '@key' => $db_entity->rets_key->value,
              '@mls' => $db_entity->rets_id->value,
              '@err' => $e->getMessage(),
            ]);
            // $this->disconnectRetsServer();
            return FALSE;
          }
          // $this->disconnectRetsServer();
        }
      }
      else {
        $item = $rets_item['result'];
      }

      $hash_fields = [];
      if ($import_media && $class->getEnableImageProcessing()) {
        $obj_mappings = $this->findDrealtyClassMappedObjects($rets_item['connection'], $rets_item['class'], $rets_item['entity_type'], $rets_item['bundle']);
        $time_fields = array_unique(array_column($obj_mappings, 'timestamp'));
        if (!empty($time_fields)) {
          $hash_fields = array_merge($time_fields, [$key_field]);
        }
      }

      if ($item && is_array($item)) {
        if (strcasecmp($db_entity->get('rets_hash')->value, $item['hash']) !== 0 || $force || !$db_entity->isPublished()) {
          // Populated mapped fields from RETS resource (excluding connected
          // RETS Objects):
          $this->set_field_data($db_entity, $item, $field_mappings, $class);
          $db_entity->set('rets_hash', $item['hash']);
          // Set as active in case it was disabled before (e.g. off market,
          // cancelled, etc.).
          $db_entity->set('status', 1);
          $db_entity->set('rets_class', $cid);

          $item_context['rets_item'] = $item;
          if ($import_media && $class->getEnableImageProcessing()) {
            // Calculate entity MediaHash value based only on EntityKey and all
            // values of the mapped Media timestamp fields.
            $media_hash = $this->calculate_hash($item, $conn_id, $cid, $hash_fields);
            if ($db_entity->get('rets_media_hash')->value !== $media_hash) {
              $rets_item = [
                'conid' => $conn_id,
                'connection' => $connection,
                'rid' => $class->get('resource'),
                'cid' => $cid,
                'key_field' => $key_field,
                'key_value' => $db_entity->get('rets_key')->value,
                'entity_type' => $entity_type,
                'bundle' => $bundle,
                'rets_media_hash' => 'update',
              ];
              $success = $this->import_single_resource_media($rets_item, $db_entity);
              if ($success) {
                // @todo when to set it? what if there's no media? look for returned code?
                $db_entity->set('rets_media_hash', $media_hash);
              }
            }
          }
          $altered = \Drupal::moduleHandler()
            ->invokeAll('drealty_import_presave', [$db_entity, $item_context]);
          if (!empty($altered['listing'])) {
            $db_entity = $altered['listing'];
          }
          $db_entity->set('changed', time());
          $db_entity->save();

          // \Drupal::moduleHandler()->invokeAll('drealty_entity_save', [
          // &$db_entity,
          // $item_context,
          // ]);
        }
        else {
          // Skipping this item.
          $this->MessageErrorLogger("Item has not changed. Skipping item. MLS Key = @rets_key, Drupal ID = @id, type @entity_type]", [
            "@entity_type" => $entity_type . ' || ' . $bundle,
            "@rets_key" => $item[$key_field],
            "@id" => $db_entity->id(),
          ], 'both', 'notice');
          return 'skip';
        }

        return TRUE;
      }
    }
    catch (\Exception $e) {
      return FALSE;
    }

    return FALSE;
  }

  /**
   * Set basic RETS mapped fields values.
   */
  protected function set_field_data(&$item, $rets_item, $field_mappings, $class) {
    foreach ($field_mappings as $local_field_id => $field_sett) {
      if (!in_array($local_field_id, ['something'])) {
        unset($v, $field_instance, $map_item, $rets_val, $field_value);
        $field_instance = FieldConfig::loadByName($item->getEntityTypeId(), $item->bundle(), $local_field_id);
        $field_cardinality = $field_instance->getFieldStorageDefinition()
          ->getCardinality();
        $field_value = [];
        $item->set($local_field_id, NULL);
        if (strpos($field_sett['field_type'], 'list_') !== FALSE) {
          foreach ($field_sett['drealty_mapping'] as $map_item) {
            $rets_val = $rets_item[$map_item['field']];
            if (!empty($rets_val) && !empty($map_item['rets_field_type']) && $map_item['rets_field_type'] == 'LookupMulti') {
              if (strpos($rets_val, ',') !== FALSE) {
                $rets_val = explode(',', $rets_val);
                foreach ($rets_val as $v) {
                  $field_value[][$map_item['property']] = $v;
                }
              }
              else {
                $field_value[$map_item['property']] = $rets_val;
              }
            }
            elseif (!empty($rets_val)) {
              $field_value[$map_item['property']] = $rets_val;
            }
          }
        }
        // Ensure handling for Address type fields:
        elseif (strpos($field_sett['field_type'], 'address') !== FALSE) {
          foreach ($field_sett['drealty_mapping'] as $map_item) {
            if (!empty($map_item['field']) && is_string($map_item['field']) && !empty($rets_item[$map_item['field']])) {
              $rets_val = $rets_item[$map_item['field']];
            }
            elseif (!empty($map_item['field']) && is_array($map_item['field'])) {
              $rets_val = [];
              foreach ($map_item['field'] as $v) {
                $rets_val[] = $rets_item[$v];
              }
              $rets_val = implode(' ', array_filter($rets_val));
            }
            $field_value[0][$map_item['property']] = $rets_val;
          }
          // Country is a must for Address field to be saved:
          $defaults = $field_instance->getDefaultValueLiteral();
          $country = $defaults ? $defaults[0]['country_code'] : 'US';
          if (empty($field_value[0]['country_code'])) {
            $field_value[0]['country_code'] = $country;
          }
        }
        elseif ($field_sett['field_type'] === 'entity_reference') {
          foreach ($field_sett['drealty_mapping'] as $map_item) {
            // "object" is set only on Media field mappings, make sure to skip it:
            if (empty($map_item['object'])) {
              $rets_val = $rets_item[$map_item['field']];

              if (!empty($rets_val)) {
                if (!empty($map_item['rets_field_type']) && $map_item['rets_field_type'] == 'LookupMulti') {
                  if (strpos($rets_val, ',') !== FALSE) {
                    $rets_val = explode(',', $rets_val);
                  }
                }
                // Make it an array to loop through later:
                $rets_val = is_array($rets_val) ? $rets_val : [$rets_val];

                $target_type = $field_instance->getSetting('target_type');
                $handler_settings = $field_instance->getSetting('handler_settings');
                $target_bundles = $handler_settings['target_bundles'];
                // $auto_create = $handler_settings['auto_create'];
                $entity_reference_storage = $this->entityTypeManager->getStorage($target_type);

                $field_value = [];
                foreach ($rets_val as $ind => $v) {
                  if ($field_cardinality < 0 || $ind < $field_cardinality) {
                    $ref_query = $entity_reference_storage->getQuery();
                    $ref_query->condition($entity_reference_storage->getEntityType()
                      ->getKey('bundle'), $target_bundles, 'IN');
                    $ref_query->condition($entity_reference_storage->getEntityType()
                      ->getKey('label'), $v, '=');
                    $ref_result = $ref_query->execute();

                    // @todo what if multiple bundles??
                    if (empty($ref_result)/* && $auto_create*/) {
                      $target_bundle = reset($target_bundles);
                      $entity_reference_values = [
                        $entity_reference_storage->getEntityType()
                          ->getKey('bundle') => $target_bundle,
                        $entity_reference_storage->getEntityType()
                          ->getKey('published') => TRUE,
                        $entity_reference_storage->getEntityType()
                          ->getKey('owner') => 1,
                        $entity_reference_storage->getEntityType()
                          ->getKey('label') => $v,
                      ];
                      $ref_entity = $entity_reference_storage->create($entity_reference_values);
                      $ref_entity->save();
                      $field_value[][$map_item['property']] = $ref_entity->id();
                    }
                    else {
                      $field_value[][$map_item['property']] = reset($ref_result);
                    }
                  }
                }
              }
            }
          }
        }
        else {
          foreach ($field_sett['drealty_mapping'] as $map_item) {
            unset($v, $rets_val);
            $rets_val = '';
            if (!empty($map_item['field']) && is_string($map_item['field']) && !empty($rets_item[$map_item['field']])) {
              $rets_val = $rets_item[$map_item['field']];
            }
            // Ensure only numeric values for numeric type fields:
            if (strpos($field_sett['field_type'], 'number') !== FALSE &&
              !is_numeric($rets_val)) {
              continue;
            }
            // Ensure only 1|0 values for boolean type fields:
            elseif (strpos($field_sett['field_type'], 'boolean') !== FALSE) {
              $rets_val = in_array($rets_val, [
                'true',
                'True',
                'TRUE',
                'yes',
                'Yes',
                'YES',
                'y',
                'Y',
                '1',
                'on',
                'On',
                'ON',
                TRUE,
                1,
              ], TRUE) ? 1 : 0;
            }
            // Ensure handling for Geofield type fields:
            elseif (strpos($field_sett['field_type'], 'geofield') !== FALSE) {
              // This can be Lat/Lon pair value to grab as "POINT(lon lat)":
              if (!empty($map_item['field2']) && is_numeric($rets_item[$map_item['field2']]) && is_numeric($rets_item[$map_item['field']])) {
                $rets_val = "POINT (" . $rets_item[$map_item['field2']] . " " . $rets_item[$map_item['field']] . ")";
                $map_item['property'] = 'value';
              }
            }

            $field_value[$map_item['property']] = $rets_val;
            // For rich text fields add Format value:
            if (strpos($field_sett['field_type'], 'text') !== FALSE) {
              $field_value['format'] = filter_default_format();
            }
          }
        }

        if (!empty($field_value)) {
          $item->set($local_field_id, $field_value);
        }
      }
    }
  }

  /**
   * Set media mapped fields values.
   */
  protected function set_media_field_data($media_field_items, &$db_entity) {
    foreach ($media_field_items as $item) {
      $field_instance = FieldConfig::loadByName($item['entity_type'], $item['bundle'], $item['field_name']);
      $field_cardinality = $field_instance->getFieldStorageDefinition()
        ->getCardinality();
      $db_entity->set($item['field_name'], NULL);

      if (preg_match('~image|file$~', $item['field_type'])) {
        try {
          // Find source field on media entity:
          if (stripos($item['field_type'], 'media/') !== FALSE) {
            $media_source_field = MediaType::load($item['media_bundle'])
              ->getSource()->getConfiguration()['source_field'];
            if (!empty($media_source_field)) {
              $field_instance = FieldConfig::loadByName('media', $item['media_bundle'], $media_source_field);
            }
          }
          $destination = (!empty($field_instance->getSetting('file_directory'))) ? 'public://' . $field_instance->getSetting('file_directory') : 'public://media-mls';
          if (count($item['items']) && $destination) {
            $type = explode('/', $item['items'][0]->getContentType());
            $destination .= ($type && !empty($type[0])) ? '/' . $type[0] . '/' . $item['items'][0]->getContentId() : '/' . $item['items'][0]->getContentId();
          }
          $destination = PlainTextOutput::renderFromHtml(\Drupal::token()
            ->replace($destination, [], ['clear' => TRUE]));
          $new_dir = \Drupal::service('file_system')
            ->prepareDirectory($destination, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
          if (!$new_dir) {
            $this->logger->error('Could not create "%dir" directory for RETS Media', ['%dir' => $destination]);
          }

          if ($new_dir && isset($item['items']) && count($item['items'])) {
            foreach ($item['items'] as $ind => $media) {
              if ($field_cardinality < 0 || $ind < $field_cardinality) {
                $file = FALSE;
                $type = explode('/', $media->getContentType());
                if (stripos($item['field_type'], 'image') !== FALSE) {
                  $filename = $media->getContentId() . '-' . $media->getObjectId() . '.' . $type[1];
                }
                else {
                  $filename = $media->getContentId() . '-' . rawurlencode($media->getContentDescription()) . '.' . $type[1];
                }
                $filepath = !empty($destination) ? $destination . '/' . $filename : 'public://media-mls/' . $filename;
                if (strlen($media->getContent()) > 170 && $item['format'] == 'data') {
                  // Ensure that there is enough data to actually make a file.
                  $file = \Drupal::service('file.repository')
                    ->writeData($media->getContent(), $filepath, FileSystemInterface::EXISTS_REPLACE);
                }
                elseif (!empty($media->getLocation()) && $item['format'] == 'url') {
                  $file = system_retrieve_file($media->getLocation(), $filepath, TRUE, FileSystemInterface::EXISTS_REPLACE);
                }

                if ($file) {
                  if (stripos($item['field_type'], 'image') !== FALSE) {
                    $file->alt = 'Listing #' . $db_entity->get('rets_id')->value . ': image #' . $media->getObjectId();
                    $file->title = !empty($media->getContentDescription()) ? $media->getContentDescription() : $file->alt;
                  }
                  else {
                    $file->description = $media->getContentDescription();
                    $file->display = 1;
                  }
                  $file->save();
                }

                if ($file instanceof FileInterface) {

                  if (in_array($item['field_type'], ['image', 'file'])) {
                    $db_entity->{$item['field_name']}->appendItem((array) $file);
                  }
                  elseif (stripos($item['field_type'], 'media/') !== FALSE) {
                    $usage = \Drupal::service('file.usage')->listUsage($file);
                    if (isset($usage['file']['media'])) {
                      $media_id = array_key_first($usage['file']['media']);
                    }
                    else {
                      $media = Media::create(
                        [
                          'bundle' => !empty($item['media_bundle']) ? $item['media_bundle'] : 'image',
                          'uid' => \Drupal::currentUser()->id(),
                          'langcode' => \Drupal::languageManager()
                            ->getDefaultLanguage()
                            ->getId(),
                          'status' => 1,
                          'name' => 'RETS Media: ' . $file->getFilename(),
                          $media_source_field => [
                            'target_id' => $file->id(),
                            'alt' => 'File ' . $file->alt,
                            'title' => $file->title,
                          ],
                        ]
                      );
                      $media->save();
                      $media_id = $media->id();
                    }
                    if (!empty($media_id)) {
                      $db_entity->{$item['field_name']}->appendItem(['target_id' => $media_id]);
                    }
                  }
                }
                else {
                  // @todo log failure
                }
              }
              else {
                break;
              }
            }
          }
        }
        catch (\Exception $e) {
          // $this->MessageErrorLogger('DRealty Error setting the values from Media Objects: ' . $e->getMessage(), array(), 'both', WATCHDOG_ERROR);
        }
      }
      else {
        try {
          switch ($item['field_type']) {
            case 'text_with_summary':
            case 'text_long':
              if (isset($item['items']) && count($item['items'])) {
                foreach ($item['items'] as $ind => $media) {
                  if ($field_cardinality < 0 || $ind < $field_cardinality) {
                    $db_entity->{$item['field_name']}->appendItem(['value' => (!empty($media['Data']) ? $media['Data'] : NULL)]);
                  }
                }
              }
              break;

            case 'text':
              if (isset($item['items']) && count($item['items'])) {
                foreach ($item['items'] as $ind => $media) {
                  if (!empty($media['Location']) && ($field_cardinality < 0 || $ind < $field_cardinality)) {
                    $db_entity->{$item['field_name']}->appendItem(['value' => $media['Location']]);
                    // $db_entity->{$item['field_name']}[LANGUAGE_NONE][]['value'] = $media['Location'];
                  }
                }
              }
              break;

            case 'link':
              if (isset($item['items']) && count($item['items'])) {
                foreach ($item['items'] as $ind => $media) {
                  if ($field_cardinality < 0 || $ind < $field_cardinality) {
                    $sett = $field_instance->getSettings();
                    $link_title = ($sett['title'] == 'value') ? $sett['title_value'] : $sett['title_value'];
                    if (!empty($media['Location'])) {
                      $db_entity->{$item['field_name']}->appendItem([
                        'url' => $media['Location'],
                        'title' => !empty($media['Content-Description']) ? $media['Content-Description'] : (empty($link_title) ? 'Media URL' : $link_title),
                        'attributes' => serialize(['title' => !empty($sett['attributes']['title']) ? $sett['attributes']['title'] : (empty($media['Content-Description']) ? 'Media URL' : $media['Content-Description'])]),
                      ]);
                    }
                  }
                }
              }
              break;
          }
        }
        catch (\Exception $e) {
          // $this->MessageErrorLogger('DRealty Error setting the values from Media Objects: ' . $e->getMessage(), [], 'both', WATCHDOG_ERROR);
        }
      }
    }
  }

  /**
   * Initiate processing of Media for each Resource class.
   */
  public function import_media() {
    $media_queue = \Drupal::queue('drealty_media');
    $total = $media_queue->numberOfItems();
    $count = 1;
    $connections = $this->entityTypeManager
      ->getStorage('drealty_connection')
      ->loadByProperties(['status' => TRUE]);

    while ($queue_item = $media_queue->claimItem()) {
      $rets_item = $queue_item->data;
      $this->MessageErrorLogger("START Processing of Media Items for @entity_type with KEY @id. [@count of @total]", [
        "@entity_type" => $rets_item['entity_type'],
        "@id" => $rets_item['key_value'],
        "@count" => $count,
        "@total" => $total,
      ], 'drush', 'notice');

      $rets_item['connection'] = $connections[$rets_item['conid']];

      if (!empty($connections[$rets_item['conid']]) && $this->connectRetsServer($connections[$rets_item['conid']])) {
        // If ($this->dc->is_connected() || $this->dc->connect($rets_item['conid'])) {.
        if ($success = $this->import_single_resource_media($rets_item)) {
          $this->MessageErrorLogger("Processed Media Items for @entity_type with KEY @id. [@count of @total]", [
            "@entity_type" => $rets_item['entity_type'],
            "@id" => $rets_item['key_value'],
            "@count" => $count,
            "@total" => $total,
          ], 'drush', 'info');
          $media_queue->deleteItem($queue_item);
        }
        else {
          // Skipping this item.
          $this->MessageErrorLogger("Was not able to process the Media for @entity_type with KEY @id. Skipping item. [@count of @total]", [
            "@entity_type" => $rets_item['entity_type'],
            "@id" => $rets_item['key_value'],
            "@count" => $count,
            "@total" => $total,
          ], 'drush', 'notice');
          $media_queue->deleteItem($queue_item);
        }
        $count++;
      }
    }

    return TRUE;
  }

  /**
   * Import all of the Media Objects data from RETS and assign to entity fields.
   *
   * @param array $rets_item
   *   Array of field values retrieved from RETS.
   * @param mixed $db_entity
   *   Drupal entity (node, etc.) that holds RETS data.
   */
  public function import_single_resource_media($rets_item, &$db_entity = FALSE) {
    $mapped_fields = [];
    $media_field_items = [];

    if (!$db_entity) {
      $save_entity = TRUE;
    }

    if (!$db_entity && $entity_exists = $this->entityExists($rets_item['entity_type'], $rets_item['key_value'], NULL, $rets_item['cid'], $rets_item['bundle'])) {
      $db_entity = $entity_exists;
    }
    elseif (!$db_entity && !$entity_exists) {
      $this->MessageErrorLogger("No Property exists in the database with the Key value @key.", ['@key' => $rets_item['key_value']], 'drush', 'warning');
      return FALSE;
    }

    if ($db_entity) {
      if (empty($db_entity->get('rets_media_hash')->value) || strcasecmp($db_entity->get('rets_media_hash')->value, $rets_item['rets_media_hash']) !== 0) {
        $mappings = $this->findDrealtyClassMappedObjects($rets_item['conid'], $rets_item['cid'], $rets_item['entity_type'], $rets_item['bundle']);
        if ($mappings) {
          foreach ($mappings as $f_name => $f_val) {
            $mapped_fields[$f_name] = [
              'field_name' => $f_name,
              'format' => $f_val['format'],
              'field_type' => $f_val['field_api_type'],
              'obj_type' => $f_val['type'],
              'media_bundle' => $f_val['media_bundle'],
            ];
          }
        }

        // ($this->dc->is_connected() || (!$this->dc->is_connected() && $this->dc->connect($rets_item['conid'])))) {.
        if (count($mapped_fields) && $this->connectRetsServer($rets_item['connection'])) {
          foreach ($mapped_fields as $m => $v) {
            $format = ($v['format'] == 'data') ? 0 : 1;

            // If RETS Object mapping was set through RETS search query:
            if (isset($mappings[$v['field_name']]->data['media_resource'])) {
              // @todo fix this block for D8
              $field = db_select('drealty_field_mappings', 'm')
                ->fields('m', ['field_name'])
                ->condition('conid', $rets_item['conid'])
                ->condition('cid', $rets_item['cid'])
                ->condition('rid', $rets_item['rid'])
                ->condition('systemname', $mappings[$v['field_name']]->data['resource_field'])
                ->execute()->fetchField();

              if ($field && isset($db_entity->{$field}) && !empty($db_entity->{$field})) {
                $val = is_array($db_entity->{$field}) ? $db_entity->{$field}[LANGUAGE_NONE][0]['value'] : $db_entity->{$field};
                $media_resource = $this->dm->FetchResource($mappings[$v['field_name']]->data['media_resource']);
                $classes = $this->dm->FetchClasses($rets_item['conid'], $media_resource);
                $query = "({$mappings[$v['field_name']]->data['media_resource_field']}=$val)";
                if (!empty($mappings[$v['field_name']]->data['query'])) {
                  $query .= ',' . $mappings[$v['field_name']]->data['query'];
                }

                $options = [
                  'Format' => 'COMPACT-DECODED',
                  'Limit' => "NONE",
                  'Count' => '0',
                ];

                foreach ($classes as $c) {
                  $search = $this->SearchQuery($media_resource->systemname, $c->systemname, "$query", $options);
                  if (!$search) {
                    // @todo log empty result ?
                  }
                  elseif ($search && $this->dc->rets->NumRows() > 0) {
                    while ($entity = $this->dc->rets->FetchRow($search)) {
                      $rets_media_items[] = $entity;
                    }
                  }
                }
                if ($rets_media_items) {
                  // @todo convert:
                  \Drupal::moduleHandler()
                    ->invokeAll('drealty_media_entity_process', [
                      'media' => $rets_media_items,
                      'rets_item' => $rets_item,
                      'fieldname' => $v['field_name'],
                    ]);
                }
              }
            }
            // Else RETS Object mapping was set through RETS GetObject request:
            elseif ($results = $this->rets->GetObject($rets_item['rid'], $v['obj_type'], $rets_item['key_value'], '*', $format)) {
              $i = 0;
              $obj_first = $results->first();

              // Does this represent some kind of error?
              if ($obj_first->isError()) {
                // Returns a \PHRETS\Models\RETSError.
                $err = $obj_first->getError();
              }

              if ($results->count() == 1 && isset($err) && $err->getCode() == 20403) {
                continue;
              }
              else {
                $media_field_items[$m]['class_id'] = $rets_item['cid'];
                $media_field_items[$m]['entity_type'] = $rets_item['entity_type'];
                $media_field_items[$m]['bundle'] = $rets_item['bundle'];
                $media_field_items[$m] += $v;

                $results = $results->toArray();

                foreach ($results as $media) {
                  if (empty($media->getError())) {
                    $media_field_items[$m]['items'][$i] = $media;
                  }
                  $i++;
                }
              }
            }
          }
          // $this->disconnectRetsServer();
        }

        if (count($media_field_items)) {
          $this->set_media_field_data($media_field_items, $db_entity);
          $db_entity->set('rets_media_hash', $rets_item['rets_media_hash']);
          if (isset($save_entity)) {
            $db_entity->set('changed', time());
            $db_entity->save();
            // $this->MessageErrorLogger("Saved entity.", [], 'drush', 'notice');
          }
          unset($rets_item, $media_field_items, $db_entity, $rets_media_items);
          return TRUE;
        }
        elseif (isset($rets_media_items)) {
          $this->MessageErrorLogger("Media items have been found. Hook drealty_media_entity_process() needs to be implemented for processing.", [], 'drush', 'notice');
          unset($rets_item, $media_field_items, $db_entity, $rets_media_items);
          return FALSE;
        }
        else {
          $this->MessageErrorLogger("No Media items have been found. Nothing to process.", [], 'drush', 'notice');
          unset($rets_item, $media_field_items, $db_entity, $rets_media_items);
          return FALSE;
        }
      }
      else {
        $this->MessageErrorLogger("Media Hash did not change. Skipping the item.", [], 'drush', 'notice');
        unset($rets_item, $media_field_items, $db_entity, $rets_media_items);
        return FALSE;
      }
    }
  }

  /**
   * Message Logging for both - CLI and UI invocation.
   */
  public function MessageErrorLogger($msg, $options = [], $storage = 'both', $error_type = 'error') {
    if (PHP_SAPI == 'cli' && function_exists('drush_log') && in_array($storage, ['drush'])) {
      drush_log(t($msg, $options), $error_type);
      // $this->logger_drush->{$error_type}(t($msg, $options));
    }
    if ($storage !== 'drush') {
      $this->logger->{$error_type}(t($msg, $options));
    }
  }

  /**
   * EntityFieldQuery to check if RETS data matching entity exists in the database.
   *
   * @param string $entity_type
   *   Drupal entity type (e.g. "node").
   * @param string $key
   *   RETS unique id.
   * @param string $mls
   *   RETS MLS ID.
   * @param string $cid
   *   Drupal ID for RETS Class entity.
   * @param string $bundle
   *   Drupal entity bundle.
   */
  public function entityExists($entity_type, $key = NULL, $mls = NULL, $cid = NULL, $bundle = NULL) {
    if (!$key && !$mls) {
      return FALSE;
    }
    $db_entity = FALSE;

    $query = \Drupal::entityQuery($entity_type);
    // ->accessCheck(FALSE);
    if (!empty($key)) {
      $query->condition('rets_key', $key);
    }
    if (!empty($mls)) {
      $query->condition('rets_id', $mls);
    }
    if (!empty($cid)) {
      $query->condition('rets_class', $cid);
    }
    if (!empty($bundle)) {
      $query->condition('type', $bundle);
    }
    $entities_find = $query->accessCheck(FALSE)->execute();
    if (!empty($entities_find)) {
      $db_entities = $this->entityTypeManager->getStorage($entity_type)
        ->loadMultiple($entities_find);
      $db_entity = reset($db_entities);
    }

    return $db_entity;
  }

  /**
   * Get all Drupal fields settings that have RETS mapping set up for specific class.
   */
  public function findDrealtyMappedFields($conn_id, $cid = NULL, $entity_type = NULL, $bundle = NULL) {
    $result = [];
    if (!$conn_id) {
      return FALSE;
    }
    if (stripos($cid, '.') !== FALSE) {
      $cid = str_replace('.', '__', $cid);
    }
    $query = \Drupal::entityQuery('field_config');
    $query->condition('third_party_settings.drealty.mapping.' . $conn_id, '', 'IS NOT NULL');
    if (!empty($cid)) {
      $query->condition('third_party_settings.drealty.mapping.' . $conn_id . '.' . $cid, '', 'IS NOT NULL');
    }
    if (!empty($entity_type)) {
      $query->condition('entity_type', $entity_type);
    }
    if (!empty($bundle)) {
      $query->condition('bundle', $bundle);
    }
    $found = $query->accessCheck(FALSE)->execute();
    if (!empty($found)) {
      $res = $this->entityTypeManager
        ->getStorage('field_config')
        ->loadMultiple($found);
      foreach ($res as $r) {
        $arr = $r->toArray();
        if (!empty($cid)) {
          $arr['drealty_mapping'] = NestedArray::getValue($arr['third_party_settings'], [
            'drealty',
            'mapping',
            $conn_id,
            $cid,
          ]);
        }
        else {
          $arr['drealty_mapping'] = NestedArray::getValue($arr['third_party_settings'], [
            'drealty',
            'mapping',
            $conn_id,
          ]);
        }
        unset($arr['third_party_settings']);
        if (!empty($arr['drealty_mapping'])) {
          $result[$r->getName()] = $arr;
        }
      }
    }
    return empty($result) ? FALSE : $result;
  }

  /**
   * Get a list of RETS Objects from DB config that can exist for this class in RETS.
   */
  public function findDrealtyClassObjects($conn_id, $resource_id, $cid) {
    $result = [];
    if (!$conn_id && !$resource_id && !$cid) {
      return FALSE;
    }
    if (stripos($cid, '.') !== FALSE) {
      $cid = str_replace('.', '__', $cid);
    }
    $query = \Drupal::entityQuery('drealty_connection')
      ->condition('id', $conn_id);
    $query->condition('resources.' . $resource_id . '.classes.' . $cid . '.objects', '', 'IS NOT NULL');
    $found = $query->accessCheck(FALSE)->execute();

    if (!empty($found)) {
      $res = $this->entityTypeManager
        ->getStorage('drealty_connection')
        ->loadMultiple($found);
      foreach ($res as $r) {
        $arr = $r->toArray();
        $result = NestedArray::getValue($arr['resources'], [
          $resource_id,
          'classes',
          $cid,
          'objects',
        ]);
      }
    }
    return empty($result) ? FALSE : $result;
  }

  /**
   * Find all RETS Objects that are mapped to fields of the entity with specific RETS class mapping.
   */
  public function findDrealtyClassMappedObjects($conn_id, $cid = NULL, $entity_type = NULL, $bundle = NULL) {
    $result = [];
    if (!$conn_id) {
      return FALSE;
    }
    if (stripos($cid, '.') !== FALSE) {
      $cid = str_replace('.', '__', $cid);
    }
    $query = \Drupal::entityQuery('field_config');
    $query->condition('third_party_settings.drealty.mapping.' . $conn_id, '', 'IS NOT NULL');
    if (!empty($cid)) {
      $query->condition('third_party_settings.drealty.mapping.' . $conn_id . '.' . $cid . '.mls_object', '', 'IS NOT NULL');
    }
    if (!empty($entity_type)) {
      $query->condition('entity_type', $entity_type);
    }
    if (!empty($bundle)) {
      $query->condition('bundle', $bundle);
    }
    $found = $query->accessCheck(FALSE)->execute();
    if (!empty($found)) {
      $res = $this->entityTypeManager
        ->getStorage('field_config')
        ->loadMultiple($found);
      foreach ($res as $r) {
        $arr = $r->toArray();
        if (!empty($cid)) {
          $o = NestedArray::getValue($arr['third_party_settings'], [
            'drealty',
            'mapping',
            $conn_id,
            $cid,
            'mls_object',
          ]);
          $result[$arr['field_name']] = $o;
        }
        else {
          $o = NestedArray::getValue($arr['third_party_settings'], [
            'drealty',
            'mapping',
            $conn_id,
          ]);
          $result[$arr['field_name']] = $o;
        }
      }
    }
    return empty($result) ? FALSE : $result;
  }

  /**
   * Find all RETS Classes that are mapped for import and return RETS class mapping.
   */
  public function findClassesMappedEntities($conn_id = NULL, array $classes = []) {
    $q = \Drupal::entityQuery('drealty_class')
      ->condition('status', TRUE)
      ->condition('drealty_mapping', NULL, 'IS NOT NULL');
    if ($conn_id) {
      $q->condition('connection', $conn_id, '=');
    }
    if (!empty($classes)) {
      $q->condition('id', $classes, 'IN');
    }
    $found = $q->execute();

    $class_mappings = [];
    if (!empty($found)) {
      $classes_config = $this->entityTypeManager
        ->getStorage('drealty_class')
        ->loadMultiple($found);
      foreach ($classes_config as $class) {
        if (empty($class_mappings)) {
          $class_mappings = $class->drealty_mapping;
        }
        else {
          foreach ($class->drealty_mapping as $map_type => $map_bundles) {
            if (empty($class_mappings[$map_type])) {
              $class_mappings[$map_type] = $map_bundles;
            }
            else {
              $class_mappings[$map_type] += $map_bundles;
            }
          }
        }
      }
    }
    return $class_mappings;
  }

}
