<?php

namespace Drupal\drealty;

use Drupal\drealty\Entity\ConnectionInterface;
use Drupal\drealty\Metadata\Resource;
use Drupal\drealty\Metadata\RetsClass;

/**
 * Fetches property information from RETS server.
 */
interface RetsFetcherInterface {

  /**
   * Connect to RETS with specific connection credentials.
   *
   * @param \Drupal\drealty\Entity\ConnectionInterface $connection
   *   RETS Connection.
   *
   * @return boolean|\PHRETS\Models\Bulletin
   *   Whether the session was successfully open/started.
   */
  public function connectRetsServer(ConnectionInterface $connection);

  /**
   * Retrieve RETS resources metadata and convert into Resource entity array.
   *
   * @param \Drupal\drealty\Entity\ConnectionInterface $connection
   *   RETS Connection.
   *
   * @return \Drupal\drealty\Metadata\Resource[]
   *   Array of Metadata Resource objects.
   */
  public function getRETSResourcesStructured(ConnectionInterface $connection);

  /**
   * @param \Drupal\drealty\Entity\ConnectionInterface $connection
   *   RETS Connection.
   *
   * @return \Illuminate\Support\Collection|\PHRETS\Models\Metadata\Resource[]
   *   Array of Metadata Resource objects.
   * @throws \PHRETS\Exceptions\CapabilityUnavailable
   * @throws \PHRETS\Exceptions\MetadataNotFound
   */
  public function getRemoteResources(ConnectionInterface $connection);

  /**
   * Retrieve MLS Classes for Resource type(s).
   *
   * @param \Drupal\drealty\Entity\ConnectionInterface $connection
   *   Active DRealty Connection config entity to search on.
   * @param \Drupal\drealty\Metadata\Resource|null $resource
   *   (optional) DRealty Resource type entity (e.g. Agent, Property, etc.). If
   *   is empty, then will return all available classes for all Resources.
   */
  public function getRemoteClasses(ConnectionInterface $connection, Resource $resource = NULL);

  /**
   * Retrieve MLS Classes metadata for Resource type(s).
   *
   * @param \Drupal\drealty\Entity\ConnectionInterface $connection
   *   Active DRealty Connection config entity to search on.
   * @param \Drupal\drealty\Metadata\Resource|null $resource
   *   (optional) DRealty Resource type entity (e.g. Agent, Property, etc.). If
   *   is empty, then will return all available classes for all Resources.
   *
   * @return \Drupal\drealty\Metadata\RetsClass[]
   *   Array of Metadata Class objects.
   */
  public function getClassesMetadata(ConnectionInterface $connection, Resource $resource = NULL);

  /**
   * Retrieve field from storage in the current database.
   *
   * @param \Drupal\drealty\Entity\ConnectionInterface $connection
   *   RETS Connection.
   * @param \Drupal\drealty\Metadata\RetsClass $resource
   *   RETS Resource.
   *
   * @return \Drupal\drealty\Metadata\Field[]
   *   Array of Metadata for the fields.
   */
  public function getFields(ConnectionInterface $connection, RetsClass $resource = NULL);

  /**
   * Retrieve field from RETS server based on RETS Resource.
   *
   * @param \Drupal\drealty\Entity\ConnectionInterface $connection
   *   RETS Connection.
   * @param \Drupal\drealty\Metadata\RetsClass $resource
   *   RETS Resource.
   *
   * @return \Drupal\drealty\Metadata\Field[]
   *   Array of Metadata for the fields.
   */
  public function getRemoteFields(ConnectionInterface $connection, RetsClass $resource = NULL);

  /**
   * Perform search query.
   *
   * @param string $resource
   *   RETS Resource.
   * @param string $class
   *   RETS Class.
   * @param string $query
   *   Search query.
   * @param array $options
   *   Search options.
   * @param bool $recursive
   *   Search recursive option.
   *
   * @return \PHRETS\Models\Search\Results
   *   RETS search results data.
   * @throws \PHRETS\Exceptions\CapabilityUnavailable
   */
  public function SearchQuery($resource_id, $class_id, $query = '', array $options = [], bool $recursive = TRUE);

}
